#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, Any

from pygpt_net.core.types import (
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_CHAT,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import RenderEvent, KernelEvent
from pygpt_net.utils import trans


class Response:
    def __init__(self, window=None):
        """
        Response controller

        :param window: Window instance
        """
        super(Response, self).__init__()
        self.window = window

    def handle(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            status: bool
    ):
        """
        Handle Bridge success

        :param status: Result status
        :param context: BridgeContext
        :param extra: Extra data
        """
        ctx = context.ctx
        if not status:
            error = None
            if "error" in extra:
                error = extra.get("error")
            self.window.controller.chat.log("Bridge response: ERROR")
            if error is not None:
                self.window.ui.dialogs.alert(error)
                self.window.update_status(error)
            else:
                self.window.ui.dialogs.alert(trans('status.error'))
                self.window.update_status(trans('status.error'))
        else:
            self.window.controller.chat.log_ctx(ctx, "output")  # log
            if self.window.controller.kernel.stopped():
                return

        ctx.current = False  # reset current state
        stream = context.stream
        mode = extra.get('mode', MODE_CHAT)
        reply = extra.get('reply', False)
        internal = extra.get('internal', False)
        self.window.core.ctx.update_item(ctx)

        # fix frozen chat
        if not status:
            data = {
                "meta": ctx.meta,
            }
            event = RenderEvent(RenderEvent.TOOL_CLEAR, data)
            self.window.dispatch(event)  # hide cmd waiting
            if not self.window.controller.kernel.stopped():
                self.window.controller.chat.common.unlock_input()  # unlock input
            # set state to: error
            self.window.dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
                "id": "chat",
            }))
            return

        try:
            if mode != MODE_ASSISTANT:
                ctx.from_previous()  # append previous result if exists
                self.window.controller.chat.output.handle(
                    ctx,
                    mode,
                    stream,
                )
        except Exception as e:
            extra["error"] = e
            self.failed(context, extra)

        # post-handle, execute cmd, etc.
        self.window.controller.chat.output.post_handle(ctx, mode, stream, reply, internal)
        self.window.controller.chat.output.handle_end(ctx, mode)  # handle end.

    def begin(
            self,
            context: BridgeContext,
            extra: Dict[str, Any]
    ):
        """
        Handle Bridge begin

        :param context: BridgeContext
        :param extra: Extra data
        """
        msg = extra.get("msg", "")
        self.window.controller.chat.common.lock_input()  # lock input
        if msg:
            self.window.update_status(msg)

    def append(
            self,
            context: BridgeContext,
            extra: Dict[str, Any]
    ):
        """
        Handle Bridge append (agent mode)

        :param context: BridgeContext
        :param extra: Extra data
        """
        if self.window.controller.kernel.stopped():
            return

        ctx = context.ctx
        # at first, handle previous context (user input) if not handled yet
        prev_ctx = ctx.prev_ctx
        stream = False
        if prev_ctx.current:
            prev_ctx.current = False  # reset previous context
            self.window.core.ctx.update_item(prev_ctx)
            prev_ctx.from_previous()  # append previous result if exists
            self.window.controller.chat.output.handle(
                ctx=prev_ctx,
                mode=prev_ctx.mode,
                stream_mode=False,
            )
            self.window.controller.chat.output.post_handle(ctx=prev_ctx,
                                                           mode=prev_ctx.mode,
                                                           stream=False,
                                                           reply=False,
                                                           internal=False)

            self.window.controller.chat.output.handle_end(ctx=prev_ctx,
                                                          mode=prev_ctx.mode)  # end previous context

        # handle current step
        ctx.current = False  # reset current state
        mode = ctx.mode
        reply = ctx.reply
        internal = ctx.internal

        self.window.core.ctx.set_last_item(ctx)
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
            "stream": stream,
        }
        event = RenderEvent(RenderEvent.BEGIN, data)
        self.window.dispatch(event)

        # append step input to chat window
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
        }
        event = RenderEvent(RenderEvent.INPUT_APPEND, data)
        self.window.dispatch(event)
        self.window.core.ctx.add(ctx)
        self.window.controller.ctx.update(
            reload=True,
            all=False,
        )
        self.window.core.ctx.update_item(ctx)

        # update ctx meta
        if mode == MODE_AGENT_LLAMA and ctx.meta is not None:
            self.window.core.ctx.replace(ctx.meta)
            self.window.core.ctx.save(ctx.meta.id)
            # update preset if exists
            preset = self.window.controller.presets.get_current()
            if preset is not None:
                if ctx.meta.assistant is not None:
                    preset.assistant_id = ctx.meta.assistant
                    self.window.core.presets.update_and_save(preset)

        try:
            self.window.controller.chat.output.handle(ctx, mode, stream)
        except Exception as e:
            self.window.controller.chat.log("Output ERROR: {}".format(e))  # log
            self.window.controller.chat.handle_error(e)
            print("Error in append text: " + str(e))

        # post-handle, execute cmd, etc.
        self.window.controller.chat.output.post_handle(ctx, mode, stream, reply, internal)

        data = {
            "meta": ctx.meta,
        }
        event = RenderEvent(RenderEvent.TOOL_BEGIN, data)
        self.window.dispatch(event)  # show cmd waiting
        self.window.controller.chat.output.handle_end(ctx, mode)  # handle end.

        event = RenderEvent(RenderEvent.RELOAD)
        self.window.dispatch(event)

        # if continue reasoning
        if ctx.extra is None or (type(ctx.extra) == dict and "agent_finish" not in ctx.extra):
            self.window.update_status(trans("status.agent.reasoning"))
            self.window.controller.chat.common.lock_input()  # lock input, re-enable stop button

        if ctx.extra is not None and (type(ctx.extra) == dict and "agent_finish" in ctx.extra):
            self.window.controller.agent.llama.on_finish(ctx)  # evaluate response and continue if needed

    def end(
            self,
            context: BridgeContext,
            extra: Dict[str, Any]
    ):
        """
        Handle Bridge end

        :param context: BridgeContext
        :param extra: Extra data
        """
        msg = extra.get("msg", "")
        status = trans("status.finished")
        if msg:
            status = msg
        self.window.update_status(status)
        self.window.controller.agent.llama.on_end()
        self.window.controller.chat.common.unlock_input()  # unlock input
        # set state to: idle
        self.window.dispatch(KernelEvent(KernelEvent.STATE_IDLE, {
            "id": "chat",
        }))

    def failed(
            self,
            context: BridgeContext,
            extra: Dict[str, Any]
    ):
        """
        Handle Bridge failed

        :param context: BridgeContext
        :param extra: Extra data
        """
        err = extra.get("error") if "error" in extra else None
        self.window.controller.chat.log("Output ERROR: {}".format(err))  # log
        self.window.controller.chat.handle_error(err)
        self.window.controller.chat.common.unlock_input()  # unlock input
        print("Error in sending text: " + str(err))
        # set state to: error
        self.window.dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
            "id": "chat",
        }))

    def update_status(
            self,
            context: BridgeContext,
            extra: Dict[str, Any]
    ):
        """
        Handle Bridge evaluate

        :param context: BridgeContext
        :param extra: Extra data
        """
        msg = extra.get("msg", "")
        self.window.update_status(msg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 03:00:00                  #
# ================================================== #

class KernelDebug:
    def __init__(self, window=None):
        """
        Kernel debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'kernel'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, 'Busy:', str(self.window.controller.kernel.busy))
        self.window.core.debug.add(self.id, 'Halt:', str(self.window.controller.kernel.halt))
        self.window.core.debug.add(self.id, 'Status:', str(self.window.controller.kernel.status))
        self.window.core.debug.add(self.id, 'State:', str(self.window.controller.kernel.state))
        self.window.core.debug.add(self.id, 'Stack:', str(self.window.controller.kernel.last_stack))
        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import List

from pygpt_net.provider.agents.base import BaseAgent


class Provider:
    def __init__(self, window=None):
        """
        Agent providers

        :param window: Window instance
        """
        self.window = window
        self.agents = {}

    def get_ids(self) -> List[str]:
        """
        Get agent providers ids

        :return: agent providers ids
        """
        return list(self.agents.keys())

    def has(self, id: str) -> bool:
        """
        Check if agent exists

        :param id: agent id
        :return: True if exists
        """
        return id in self.agents

    def get(self, id: str) -> BaseAgent:
        """
        Get agent provider

        :param id: agent id
        :return: agent provider
        """
        return self.agents[id]

    def register(self, id: str, agent):
        """
        Register Agent provider

        :param id: Agent id
        :param agent: Agent provider
        """
        self.agents[id] = agent

    def get_providers(self) -> List[str]:
        """
        Get agent providers list

        :return: list of agent providers
        """
        return self.get_ids()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import os
from typing import Dict

from PySide6.QtCore import QTimer, Slot
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QFileDialog, QWidget

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.text.utils import output_clean_html, output_html2text
from pygpt_net.tools.base import BaseTool
from pygpt_net.utils import trans

from .ui.dialogs import Tool
from .ui.widgets import ToolWidget, ToolSignals


class HtmlCanvas(BaseTool):
    def __init__(self, *args, **kwargs):
        """
        HTML/JS Canvas

        :param window: Window instance
        """
        super(HtmlCanvas, self).__init__(*args, **kwargs)
        self.id = "html_canvas"
        self.dialog_id = "html_canvas"
        self.has_tab = True
        self.tab_title = "menu.tools.html_canvas"
        self.tab_icon = ":/icons/code.svg"
        self.opened = False
        self.is_edit = False
        self.auto_clear = True
        self.dialog = None
        self.is_edit = False
        self.auto_opened = False
        self.file_output = ".canvas.html"
        self.signals = ToolSignals()

    def setup(self):
        """Setup"""
        self.load_output()
        self.update()

    def on_reload(self):
        """On app profile reload"""
        self.setup()

    def update(self):
        """Update menu"""
        self.update_menu()

    def toggle_edit(self, widget: ToolWidget):
        """
        Toggle edit mode

        :param widget: Canvas widget
        """
        current = self.is_edit
        self.is_edit = not self.is_edit
        widget.edit.setVisible(self.is_edit)
        widget.output.setVisible(not self.is_edit)
        if current:
            self.set_output(widget.edit.toPlainText())
            self.save_output()

    def update_menu(self):
        """Update menu"""
        """
        if self.opened:
            self.window.ui.menu['tools.html_canvas'].setChecked(True)
        else:
            self.window.ui.menu['tools.html_canvas'].setChecked(False)
        """

    def get_current_path(self) -> str:
        """
        Get current output path

        :return: Output path
        """
        return os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)

    def get_dialog_id(self) -> str:
        """
        Get dialog ID

        :return: Dialog ID
        """
        return self.dialog_id

    def set_output(self, output: str):
        """
        Set output HTML

        :param output: Output HTML code
        """
        path = self.get_current_path()
        with open(path, "w", encoding="utf-8") as f:
            f.write(output)
        if os.path.exists(path):
            self.signals.reload.emit(path)
        self.signals.update.emit(output)

    def reload_output(self):
        """Reload output data"""
        self.load_output()
        self.update()

    def get_output(self) -> str:
        """
        Get current HTML output

        :return: Output data
        """
        path = self.get_current_path()
        data = ""
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                data = f.read()
        return data

    def load_output(self):
        """Load output data from file"""
        path = self.get_current_path()
        data = ""
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                data = f.read()
        self.set_output(data)

    def save_output(self):
        """Save output data to file"""
        path = self.get_current_path()
        data = self.get_output()
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)

    def clear_output(self):
        """Clear output"""
        path = self.get_current_path()
        if os.path.exists(path):
            os.remove(path)
        self.set_output("")

    def clear(self, force: bool = False):
        """
        Clear current window

        :param force: Force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='html_canvas.clear',
                id=0,
                msg=trans("html_canvas.clear.confirm"),
            )
            return
        self.clear_output()

    def clear_all(self):
        """Clear input and output"""
        self.clear_output()

    def open(self):
        """Open HTML canvas dialog"""
        if not self.opened:
            self.opened = True
            self.load_output()
            self.window.ui.dialogs.open(self.dialog_id, width=800, height=600)
            self.update()

    def auto_open(self):
        """Auto open canvas dialog"""
        if self.window.controller.ui.tabs.is_current_tool(self.id):
            return # do not open if already opened in tab
        if not self.auto_opened:
            self.auto_opened = True
            self.open()

    def open_file(self):
        """Open file dialog"""
        last_dir = self.window.core.config.get_last_used_dir()
        dialog = QFileDialog(self.window)
        dialog.setDirectory(last_dir)
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.setNameFilter("HTML files (*.html)")
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        if dialog.exec():
            path = dialog.selectedFiles()[0]
            with open(path, "r", encoding="utf-8") as f:
                data = f.read()
                self.set_output(data)
            self.open()

    def close(self):
        """Close HTML canvas dialog"""
        self.opened = False
        self.window.ui.dialogs.close(self.dialog_id)
        self.update()

    def toggle(self):
        """Toggle HTML canvas dialog open/close"""
        if self.opened:
            self.close()
        else:
            self.open()

    def show_hide(self, show: bool = True):
        """
        Show/hide HTML canvas window

        :param show: show/hide
        """
        if show:
            self.open()
        else:
            self.close()

    def get_toolbar_icon(self) -> QWidget:
        """
        Get toolbar icon

        :return: QWidget
        """
        return self.window.ui.nodes['icon.html_canvas']

    def toggle_icon(self, state: bool):
        """
        Toggle canvas icon

        :param state: State
        """
        self.get_toolbar_icon().setVisible(state)

    def get_current_output(self) -> str:
        """
        Get current output from canvas

        :return: Output data
        """
        return self.get_output()

    @Slot(str, str)
    def handle_save_as(self, text: str, type: str = 'txt'):
        """
        Handle save as signal

        :param text: Data to save
        :param type: File type
        """
        if type == 'html':
            text = output_clean_html(text)
        else:
            text = output_html2text(text)
        # fix: QTimer required here to prevent crash if signal emitted from WebEngine window
        QTimer.singleShot(0, lambda: self.window.controller.chat.common.save_text(text, type))

    def setup_menu(self) -> Dict[str, QAction]:
        """
        Setup main menu

        :return dict with menu actions
        """
        actions = {}
        actions["html_canvas"] = QAction(
            QIcon(":/icons/code.svg"),
            trans("menu.tools.html_canvas"),
            self.window,
            checkable=False,
        )
        actions["html_canvas"].triggered.connect(
            lambda: self.toggle()
        )
        return actions

    def as_tab(self, tab: Tab) -> QWidget:
        """
        Spawn and return tab instance

        :param tab: Parent Tab instance
        :return: Tab widget instance
        """
        canvas = Tool(window=self.window, tool=self)
        layout = canvas.widget.setup()
        widget = QWidget()
        widget.setLayout(layout)
        canvas.set_tab(tab)
        self.load_output()
        return widget

    def setup_dialogs(self):
        """Setup dialogs (static)"""
        self.dialog = Tool(window=self.window, tool=self)
        self.dialog.setup()

    def setup_theme(self):
        """Setup theme"""
        pass

    def get_lang_mappings(self) -> Dict[str, Dict]:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {
            'menu.text': {
                'tools.html_canvas': 'menu.tools.html_canvas',
            }
        }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any, Optional

from packaging.version import Version


class BaseProvider:
    def __init__(self, window=None):
        self.window = window
        self.id = ""
        self.type = "config"
        self.path = None
        self.path_app = None
        self.meta = None

    def attach(self, window):
        self.window = window

    def install(self):
        pass

    def patch(self, version: Version) -> bool:
        pass

    def load(self, all: bool = False) -> Optional[Dict[str, Any]]:
        pass

    def load_base(self) -> Optional[Dict[str, Any]]:
        pass

    def save(self, items: Dict[str, Any], filename: str = 'config.json'):
        pass

    def get_options(self) -> Optional[Dict[str, Any]]:
        pass

    def get_sections(self) -> Optional[Dict[str, Any]]:
        pass

    def get_version(self) -> str:
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.31 19:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from llama_index.llms.deepseek import DeepSeek
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem


class DeepseekApiLLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(DeepseekApiLLM, self).__init__(*args, **kwargs)
        self.id = "deepseek_api"
        self.type = [MODE_LLAMA_INDEX]

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        args = self.parse_args(model.llama_index)
        return DeepSeek(**args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any

from llama_index.agent.openai import OpenAIAssistantAgent as Agent

from pygpt_net.core.bridge.context import BridgeContext
from .base import BaseAgent


class OpenAIAssistantAgent(BaseAgent):
    def __init__(self, *args, **kwargs):
        super(OpenAIAssistantAgent, self).__init__(*args, **kwargs)
        self.id = "openai_assistant"
        self.mode = "assistant"  # step|plan|...

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        context = kwargs.get("context", BridgeContext())
        tools = kwargs.get("tools", [])
        verbose = kwargs.get("verbose", False)
        model = context.model
        system_prompt = context.system_prompt
        ctx = context.ctx
        thread_id = None
        assistant_id = None
        name = "Assistant"
        if (ctx is not None
                and ctx.output_name is not None
                and ctx.output_name != ""):
            name = ctx.output_name

        # get assistant_id and thread_id from ctx
        if (ctx is not None
                and ctx.meta is not None):
            if ctx.meta.assistant is not None:
                assistant_id = ctx.meta.assistant
            if ctx.meta.thread is not None:
                thread_id = ctx.meta.thread

        # get assistant_id from preset
        preset_assistant_id = context.assistant_id
        if (preset_assistant_id is not None
                and preset_assistant_id != ""):
            assistant_id = preset_assistant_id  # override assistant_id from ctx

        kwargs = {
            "tools": tools,
            "verbose": verbose,
        }
        if thread_id is not None:
            kwargs["thread_id"] = thread_id

        if assistant_id is not None:
            kwargs["assistant_id"] = assistant_id
            return Agent.from_existing(**kwargs)
        else:
            kwargs["name"] = name
            kwargs["instructions"] = system_prompt
            kwargs["openai_tools"] = [{"type": "code_interpreter"}, {"type": "file_search"}]
            kwargs["model"] = model.id
            return Agent.from_new(**kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

import os
from typing import Optional, Any

import docker
import io
import tarfile

from docker.errors import DockerException

class Docker:
    def __init__(self, plugin = None):
        self.plugin = plugin
        self.client = None
        self.container_name = "pygpt_container"
        self.image_name = "pygpt_image"
        self.initialized = False
        self.signals = None

    def get_dockerfile(self) -> str:
        """
        Get the Dockerfile

        :return: Dockerfile.
        """
        return self.plugin.get_option_value('dockerfile')

    def get_image_name(self) -> str:
        """
        Get the image name

        :return: Image name.
        """
        return self.plugin.get_option_value('image_name')

    def get_container_name(self) -> str:
        """
        Get the container name

        :return: Container name.
        """
        return self.plugin.get_option_value('container_name')

    def create_docker_context(self, dockerfile: str) -> io.BytesIO:
        """
        Create a Docker context with the specified Dockerfile content.

        :param dockerfile: Dockerfile content.
        :return: Docker context.
        """
        tar_stream = io.BytesIO()
        with tarfile.open(fileobj=tar_stream, mode='w') as tar:
            dockerfile_info = tarfile.TarInfo('Dockerfile')
            dockerfile_data = dockerfile.encode('utf-8')
            dockerfile_info.size = len(dockerfile_data)
            tar.addfile(dockerfile_info, io.BytesIO(dockerfile_data))
        tar_stream.seek(0)
        return tar_stream

    def is_image(self) -> bool:
        """
        Check if the Docker image exists.

        :return: True if the image exists.
        """
        client = self.get_docker_client()
        try:
            client.images.get(self.get_image_name())
            return True
        except docker.errors.ImageNotFound as e:
            print(e)
            return False

    def build_image(self):
        """Build the Docker image."""
        client = self.get_docker_client()
        context = self.create_docker_context(self.get_dockerfile())
        self.log("Please wait... Building the Docker image...")
        image, logs = client.images.build(
            fileobj=context,
            custom_context=True,
            rm=True,
            tag=self.get_image_name(),
        )
        for chunk in logs:
            if 'stream' in chunk:
                self.log(chunk['stream'].strip())

    def prepare_local_data_dir(self):
        """ Prepare the local data directory."""
        local_data_dir = self.get_local_data_dir()
        try:
            os.makedirs(local_data_dir)
        except FileExistsError:
            pass

    def get_docker_client(self) -> docker.DockerClient:
        """
        Get the Docker client.

        :return: Docker client.
        """
        return docker.from_env()

    def end(self, all: bool = False):
        """
        Stop all.

        :param all: Stop the container as well.
        """
        if all:
            self.stop_container(self.get_container_name())

    def stop_container(self, name: str):
        """
        Stop the Docker container.

        :param name: Container name.
        """
        client = self.get_docker_client()
        try:
            container = client.containers.get(name)
            container.stop()
            container.remove()
        except docker.errors.NotFound:
            self.log(f"Container '{name}' not found.")

    def create_container(self, name: str):
        """
        Create the Docker container.

        :param name: Container name.
        """
        client = self.get_docker_client()
        image_name = self.get_image_name()
        entrypoint = self.get_entrypoint()
        volumes = self.get_volumes()
        ports = self.get_ports()

        try:
            container = client.containers.get(name)
            container.reload()
            if container.status == 'running':
                pass
            else:
                print(f"Container '{name}' is not running. Starting it.")
                container.remove()
                container = client.containers.create(
                    image=image_name,
                    name=name,
                    volumes=volumes,
                    ports=ports,
                    tty=True,
                    stdin_open=True,
                    command=entrypoint,
                )
                container.start()
        except docker.errors.NotFound:
            print(f"Container '{name}' not found. Creating a new one.")
            container = client.containers.create(
                image=image_name,
                name=name,
                volumes=volumes,
                ports=ports,
                tty=True,
                stdin_open=True,
                command=entrypoint,
            )
            container.start()
        except Exception as e:
            self.log(f"Error creating container: {e}")

    def restart_container(self, name: str):
        """
        Restart the Docker container.

        :param name: Container name.
        """
        client = self.get_docker_client()
        image_name = self.get_image_name()
        entrypoint = self.get_entrypoint()
        volumes = self.get_volumes()
        ports = self.get_ports()

        try:
            container = client.containers.get(name)
            container.reload()
            status = container.status
            print(f"Container '{name}' status: {status}")

            if status == 'running':
                print(f"Stopping and starting container '{name}'...")
                container.stop()
                container.wait()
                container.reload()

            elif status == 'paused':
                print(f"Resuming and starting container '{name}'...")
                container.unpause()
                container.stop()
                container.wait()
                container.reload()

            elif status in ['exited', 'created']:
                print(f"Container '{name}' is in state '{status}'. Starting it.")

            elif status == 'restarting':
                print(f"Container '{name}' is restarting. Waiting...")
                container.wait()
                container.reload()

            elif status == 'removing':
                print(f"Container '{name}' is being removed. Waiting...")
                container.wait()
                container = None

            elif status == 'dead':
                print(f"Container '{name}' is dead. Removing and creating a new one.")
                container.remove()
                container = None

            else:
                print(f"Unknown container status: {status}. Removing and creating a new one.")
                container.remove()
                container = None

            if container:
                print(f"Starting container '{name}'...")
                try:
                    container.start()
                    container.reload()
                    if container.status != 'running':
                        print(f"Container '{name}' did not start correctly. Status: {container.status}")
                        print(f"Removing and creating a new container '{name}'...")
                        container.remove()
                        container = None
                except Exception as e:
                    print(f"Error starting container '{name}': {e}")
                    print(f"Removing and creating a new container '{name}'...")
                    container.remove()
                    container = None

            if not container:
                print(f"Creating a new container '{name}'...")
                container = client.containers.create(
                    image=image_name,
                    name=name,
                    volumes=volumes,
                    ports=ports,
                    tty=True,
                    stdin_open=True,
                    command=entrypoint,  # 'running'
                )
                container.start()
                container.reload()
                if container.status != 'running':
                    print(f"Container '{name}' did not start correctly. Status: {container.status}")
                else:
                    print(f"Container '{name}' started successfully.")

        except docker.errors.NotFound:
            print(f"Container '{name}' not found. Creating a new one.")
            container = client.containers.create(
                image=image_name,
                name=name,
                volumes=volumes,
                ports=ports,
                tty=True,
                stdin_open=True,
                command=entrypoint,  # 'running'
            )
            container.start()
            container.reload()
            if container.status != 'running':
                print(f"Container '{name}' did not start correctly. Status: {container.status}")
            else:
                print(f"Container '{name}' started successfully.")
        except Exception as e:
            print(f"Error restarting container '{name}': {e}")

    def restart(self):
        """Restart the Docker container."""
        self.restart_container(self.get_container_name())

    def get_volumes(self) -> dict:
        """
        Get the volumes mappings.

        :return: Volumes mappings.
        """
        workdir = self.get_local_data_dir()
        config = self.plugin.get_option_value('docker_volumes')
        data = {}
        for item in config:
            if item['enabled']:
                host_dir = item['host'].format(workdir=workdir)
                data[host_dir] = {
                    'bind': item['docker'],
                    'mode': 'rw',
                }
        return data

    def get_ports(self) -> dict:
        """
        Get the ports mappings.

        :return: Ports mappings.
        """
        config = self.plugin.get_option_value('docker_ports')
        data = {}
        for item in config:
            if item['enabled']:
                docker_port = item['docker']
                try:
                    host_port = int(item['host'])
                except ValueError:
                    print("WARNING: Invalid host port number: {}. "
                          "Please provide a valid port number as integer value".format(item['host']))
                    continue
                if "/" not in docker_port:
                    docker_port = f"{docker_port}/tcp"
                data[docker_port] = host_port
        return data

    def get_entrypoint(self) -> str:
        """
        Get the Docker entrypoint.

        :return: Docker entrypoint command.
        """
        return self.plugin.get_option_value('docker_entrypoint')

    def execute(self, cmd: str) -> Optional[bytes]:
        """
        Execute command in Docker container.

        :param cmd: Command to execute
        :return: Response
        """
        client = self.get_docker_client()
        name = self.get_container_name()

        # at first, check for image
        if not self.is_image():
            self.build_image()

        # run the container
        try:
            self.create_container(name)
            container = client.containers.get(name)
            result = container.exec_run(
                cmd,
                stdout=True,
                stderr=True,
            )
            tmp = result.output.decode("utf-8")
            response = tmp.encode("utf-8")
        except Exception as e:
            self.log(f"Error running container: {e}")
            response = str(e).encode("utf-8")
        return response

    def get_local_data_dir(self) -> str:
        """
        Get the local data directory.

        :return: Local data directory.
        """
        return self.plugin.window.core.config.get_user_dir("data")

    def is_docker_installed(self) -> bool:
        """
        Check if Docker is installed

        :return: True if installed
        """
        try:
            if self.client is None:
                client = docker.from_env()
                client.ping()
            return True
        except DockerException:
            return False

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    def log(self, msg: Any):
        """
        Log the message.

        :param msg: Message to log.
        """
        print(msg)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional

from .base import BaseEvent


class RenderEvent(BaseEvent):
    """Events used for rendering"""
    BEGIN = "render.begin"
    END = "render.end"
    RELOAD = "render.reload"
    RESET = "render.reset"
    PREPARE = "render.prepare"

    STREAM_BEGIN = "render.stream.begin"
    STREAM_APPEND = "render.stream.append"
    STREAM_END = "render.stream.end"

    CLEAR = "render.clear"
    CLEAR_ALL = "render.clear.all"
    CLEAR_INPUT = "render.clear.input"
    CLEAR_OUTPUT = "render.clear.output"

    INPUT_APPEND = "render.input.append"
    OUTPUT_APPEND = "render.output.append"
    CTX_APPEND = "render.ctx.append"

    EXTRA_BEGIN = "render.extra.begin"
    EXTRA_APPEND = "render.extra.append"
    EXTRA_END = "render.extra.end"

    ON_LOAD = "render.on.load"
    ON_PAGE_LOAD = "render.on.page.load"
    ON_THEME_CHANGE = "render.on.theme.change"
    ON_EDIT_ENABLE = "render.on.edit.enable"
    ON_EDIT_DISABLE = "render.on.edit.disable"
    ON_TS_ENABLE = "render.on.ts.enable"
    ON_TS_DISABLE = "render.on.ts.disable"
    ON_SWITCH = "render.on.switch"

    TOOL_BEGIN = "render.tool.begin"
    TOOL_UPDATE = "render.tool.update"
    TOOL_CLEAR = "render.tool.clear"
    TOOL_END = "render.tool.end"

    ACTION_REGEN_SUBMIT = "render.action.regen.submit"
    ACTION_EDIT_SUBMIT = "render.action.edit.submit"

    def __init__(
            self,
            name: Optional[str] = None,
            data: Optional[dict] = None,
    ):
        """
        Event object class

        :param name: event name
        :param data: event data
        """
        super(RenderEvent, self).__init__(name, data)
        self.id = "RenderEvent"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.16 05:00:00                  #
# ================================================== #

from PySide6.QtCore import QObject, Signal

class BaseSignals(QObject):
    finished = Signal(object, object, dict)  # response dict, ctx, extra_data
    finished_more = Signal(list, object, dict)  # responses list, ctx, extra_data
    debug = Signal(object)
    destroyed = Signal()
    error = Signal(object)
    log = Signal(object)
    started = Signal()
    status = Signal(object)
    stopped = Signal()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 17:00:00                  #
# ================================================== #

import re
from typing import Union, Optional, Tuple, List

from PySide6.QtMultimedia import QMediaDevices
from bs4 import UnicodeDammit

from pygpt_net.provider.audio_input.base import BaseProvider as InputBaseProvider
from pygpt_net.provider.audio_output.base import BaseProvider as OutputBaseProvider

from .capture import Capture
from .whisper import Whisper


class Audio:
    def __init__(self, window=None):
        """
        Audio input/output core

        :param window: Window instance
        """
        self.window = window
        self.capture = Capture(window)
        self.whisper = Whisper(window)
        self.providers = {
            "input": {},
            "output": {},
        }
        self.last_error = None

    def get_input_devices(self) -> List[Tuple[int, str]]:
        """
        Get input devices

        :return devices list: [(id, name)]
        """
        devices = QMediaDevices.audioInputs()
        devices_list = []
        for index, device in enumerate(devices):
            dammit = UnicodeDammit(device.description())
            devices_list.append((index, dammit.unicode_markup))
        return devices_list

    def is_device_compatible(self, device_index: int) -> bool:
        """
        Check if device is compatible

        :param device_index: device index
        :return: True if compatible
        """
        import pyaudio
        rate = int(self.window.core.config.get('audio.input.rate', 44100))
        channels = int(self.window.core.config.get('audio.input.channels', 1))
        p = pyaudio.PyAudio()
        info = p.get_device_info_by_index(device_index)
        try:
            p.is_format_supported(
                rate=rate,
                input_device=info['index'],
                input_channels=channels,
                input_format=pyaudio.paInt16)
            supported = True
        except ValueError as e:
            self.last_error = str(e)
            supported = False
        p.terminate()
        return supported

    def is_registered(self, id: str, type: str = "output") -> bool:
        """
        Check if provider is registered

        :param id: provider id
        :param type: provider type
        :return: True if registered
        """
        if type in self.providers:
            return id in self.providers[type]
        return False

    def get_providers(self, type: str = "output") -> dict:
        """
        Get all providers

        :param type: provider type
        :return: providers dict
        """
        if type in self.providers:
            return self.providers[type]
        return {}

    def get_ids(self, type: str = "output") -> list:
        """
        Get all providers ids

        :param type: provider type
        :return: providers ids list
        """
        if type in self.providers:
            return list(self.providers[type].keys())
        return []

    def get(
            self,
            id: str,
            type: str = "output"
    ) -> Optional[Union[InputBaseProvider, OutputBaseProvider]]:
        """
        Get provider instance

        :param id: provider id
        :param type: provider type
        :return: provider instance
        """
        if self.is_registered(id, type):
            return self.providers[type][id]
        return None

    def register(
            self,
            provider: Union[InputBaseProvider, OutputBaseProvider],
            type: str = "output"
    ):
        """
        Register provider

        :param provider: provider instance
        :param type: provider type
        """
        id = provider.id
        self.providers[type][id] = provider

    def clean_text(self, text: str) -> str:
        """
        Clean text before send to audio synthesis

        :param text: text
        :return: cleaned text
        """
        return re.sub(r'~###~.*?~###~', '', str(text))

    def get_last_error(self) -> str:
        """
        Return last error

        :return: Error
        """
        return self.last_error



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import List

from PySide6.QtWidgets import QVBoxLayout, QWidget

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.ui.widget.textarea.output import ChatOutput

from .bag import Bag
from ...item.ctx import CtxItem


class Container:
    def __init__(self, window=None):
        """
        Context output container

        :param window: Window
        """
        self.window = window
        self.bags = {}
        self.bags[0] = Bag(window)  # always create initial bag

    def get(self, tab: Tab) -> QWidget:
        """
        Register and return output

        :param tab: Tab
        :return: Widget
        """
        # plain output
        output_plain = ChatOutput(self.window)
        output_plain.set_tab(tab)

        # web
        if self.window.core.config.get("render.engine") == "web":
            from pygpt_net.ui.widget.textarea.web import ChatWebOutput, CustomWebEnginePage
            # build output
            output_html = ChatWebOutput(self.window)
            output_html.set_tab(tab)
            output_html.setPage(
                CustomWebEnginePage(self.window, output_html)
            )
            # connect signals
            output_html.signals.save_as.connect(self.window.controller.chat.render.handle_save_as)
            output_html.signals.audio_read.connect(self.window.controller.chat.render.handle_audio_read)
        else:
            # legacy
            output_html = ChatOutput(self.window)
            output_html.set_tab(tab)

        if 'output_plain' not in self.window.ui.nodes:
            self.window.ui.nodes['output_plain'] = {}
        if 'output' not in self.window.ui.nodes:
            self.window.ui.nodes['output'] = {}

        self.window.ui.nodes['output_plain'][tab.pid] = output_plain
        self.window.ui.nodes['output'][tab.pid] = output_html

        # show/hide plain/html
        if self.window.core.config.get("render.plain") is True:
            output_plain.setVisible(True)
            output_html.setVisible(False)
        else:
            output_plain.setVisible(False)
            output_html.setVisible(True)

        # build layout
        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes['output_plain'][tab.pid])
        layout.addWidget(self.window.ui.nodes['output'][tab.pid])
        layout.setContentsMargins(0, 0, 0, 0)
        widget = self.window.core.tabs.from_layout(layout)
        return widget

    def get_active_pid(self) -> int:
        """
        Get active PID

        :return: PID
        """
        pid = self.window.controller.ui.tabs.get_current_pid()
        if pid is not None:
            return pid
        return 0  # default bag

    def get_active_tab_id(self) -> int:
        """
        Get active tab ID

        :return: Tab ID
        """
        return self.get_active_pid()

    def get_active_bag(self) -> Bag:
        """
        Get active bag

        :return: Bag
        """
        tab_id = self.get_active_tab_id()
        if tab_id not in self.bags:
            self.bags[tab_id] = Bag(self.window)  # crate new empty bag if not exists
        return self.bags[tab_id]

    def get_items(self) -> List[CtxItem]:
        """
        Get ctx items

        :return: Items
        """
        return self.get_active_bag().get_items()

    def set_items(self, items: List[CtxItem]):
        """
        Set ctx items

        :param items: Items
        """
        self.get_active_bag().set_items(items)

    def clear_items(self):
        """Clear ctx items"""
        self.get_active_bag().clear_items()

    def count_items(self) -> int:
        """
        Count ctx items

        :return: Items count
        """
        return self.get_active_bag().count_items()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 04:00:00                  #
# ================================================== #

import time

from pynput.mouse import Button, Controller as MouseController
from pynput.keyboard import Key, Controller as KeyboardController
from PySide6.QtCore import Slot, Signal
from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    screenshot = Signal(dict, object)


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.window = None
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            response = None
            try:
                if item["cmd"] in self.plugin.allowed_cmds and self.plugin.has_cmd(item["cmd"]):

                    # get mouse position
                    if item["cmd"] == "get_mouse_position":
                        response = self.cmd_mouse_get_pos(item)

                    # set mouse position
                    elif item["cmd"] == "mouse_move":
                        if self.plugin.get_option_value("allow_mouse_move"):
                            response = self.cmd_mouse_move(item)

                    # mouse click
                    elif item["cmd"] == "mouse_click":
                        if self.plugin.get_option_value("allow_mouse_click"):
                            response = self.cmd_mouse_click(item)

                    # mouse scroll
                    elif item["cmd"] == "mouse_scroll":
                        if self.plugin.get_option_value("allow_mouse_scroll"):
                            response = self.cmd_mouse_scroll(item)

                    # screenshot
                    elif item["cmd"] == "get_screenshot":
                        if self.plugin.get_option_value("allow_screenshot"):
                            response = self.cmd_make_screenshot(item)

                    # keyboard key
                    elif item["cmd"] == "keyboard_key":
                        if self.plugin.get_option_value("allow_keyboard"):
                            response = self.cmd_keyboard_key(item)

                    # keyboard type
                    elif item["cmd"] == "keyboard_type":
                        if self.plugin.get_option_value("allow_keyboard"):
                            response = self.cmd_keyboard_type(item)

                    if response:
                        responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

    def cmd_mouse_get_pos(self, item: dict) -> dict:
        """
        Get mouse position

        :param item: command item
        :return: response item
        """
        try:
            self.msg = "Mouse get position"
            self.log(self.msg)
            result = self.get_current(item)
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_mouse_move(self, item: dict) -> dict:
        """
        Set mouse position

        :param item: command item
        :return: response item
        """
        error = None
        x = 0
        y = 0
        if self.has_param(item, "x"):
            x = self.get_param(item, "x")
        elif self.has_param(item, "mouse_x"):
            x = self.get_param(item, "mouse_x")
        if self.has_param(item, "y"):
            y = self.get_param(item, "y")
        elif self.has_param(item, "mouse_y"):
            y = self.get_param(item, "mouse_y")
        try:
            mouse = MouseController()
            mouse.position = (x, y)
        except Exception as e:
            error = str(e)
            self.log("Error: {}".format(e))
        try:
            result = self.get_current(item)
            if error:
                result["error"] = error
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_mouse_click(self, item: dict) -> dict:
        """
        Mouse click

        :param item: command item
        :return: response item
        """
        button = Button.left
        num = 1
        if self.has_param(item, "button"):
            btn_name = self.get_param(item, "button")
            if btn_name == "middle":
                button = Button.middle
            elif btn_name == "right":
                button = Button.right
        if self.has_param(item, "num_clicks"):
            num = int(self.get_param(item, "num_clicks"))
        try:
            mouse = MouseController()
            mouse.click(button, num)
            result = self.get_current(item)
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_mouse_scroll(self, item: dict) -> dict:
        """
        Mouse scroll

        :param item: command item
        :return: response item
        """
        x = 0
        y = 0
        if self.has_param(item, "dx"):
            x = self.get_param(item, "dx")
        if self.has_param(item, "dy"):
            y = self.get_param(item, "dy")
        try:
            mouse = MouseController()
            mouse.scroll(x, y)
            result = self.get_current(item)
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_keyboard_key(self, item: dict) -> dict:
        """
        Keyboard key press

        :param item: command item
        :return: response item
        """
        keyboard = KeyboardController()
        error = None
        if self.has_param(item, "key"):
            key = self.get_param(item, "key")
            modifier = None
            if self.has_param(item, "modifier"):
                tmp_modifier = self.get_param(item, "modifier")
                if tmp_modifier == "ctrl":
                    modifier = Key.ctrl
                elif tmp_modifier == "alt":
                    modifier = Key.alt
                elif tmp_modifier == "shift":
                    modifier = Key.shift
                elif tmp_modifier == "cmd":
                    modifier = Key.cmd

            # autofocus on the window
            if self.plugin.get_option_value("auto_focus"):
                self.set_focus()
                time.sleep(1)  # wait for a second

            if key == "super" or key == "start":
                key = Key.cmd

            try:
                if modifier:
                    with keyboard.pressed(modifier):
                        keyboard.press(key)
                        keyboard.release(key)
                else:
                    keyboard.press(key)
                    keyboard.release(key)
            except Exception as e:
                error = str(e)
                self.log("Error: {}".format(e))

        try:
            result = self.get_current(item)
            if error:
                result["error"] = error
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_keyboard_type(self, item: dict) -> dict:
        """
        Keyboard key press

        :param item: command item
        :return: response item
        """
        keyboard = KeyboardController()
        if self.has_param(item, "text"):
            text = self.get_param(item, "text")
            modifier = None
            if self.has_param(item, "modifier"):
                tmp_modifier = self.get_param(item, "modifier")
                if tmp_modifier == "ctrl":
                    modifier = Key.ctrl
                elif tmp_modifier == "alt":
                    modifier = Key.alt
                elif tmp_modifier == "shift":
                    modifier = Key.shift
                elif tmp_modifier == "cmd":
                    modifier = Key.cmd

            # autofocus on the window
            if self.plugin.get_option_value("auto_focus"):
                self.set_focus()
                time.sleep(1)  # wait for a second

            if modifier:
                with keyboard.pressed(modifier):
                    keyboard.type(text)
            else:
                keyboard.type(text)

        try:
            result = self.get_current(item)
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_make_screenshot(self, item: dict):
        """
        Make screenshot (send signal)

        :param item: command item
        :return: response item
        """
        try:
            result = self.get_current(item)
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        # make screenshot will be handled in the main thread (in response)
        return self.make_response(item, result)

    def set_focus(self):
        """Set focus to the current mouse position"""
        try:
            mouse = MouseController()
            mouse.click(Button.left, 1)
        except Exception as e:
            pass

    def get_current(self, item: dict = None) -> dict:
        """
        Get current mouse position and screen resolution

        :param item: command item
        :return: coordinates and screen resolution
        """
        current_step = self.get_param(item, "current_step", "")
        screen = self.window.app.primaryScreen()
        size = screen.size()
        screen_x = size.width()
        screen_y = size.height()
        mouse = MouseController()
        mouse_pos_x, mouse_pos_y = mouse.position
        return {
            'current_step': current_step,
            'screen_w': screen_x,
            'screen_h': screen_y,
            'mouse_x': mouse_pos_x,
            'mouse_y': mouse_pos_y,
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.31 19:00:00                  #
# ================================================== #

import os
from typing import Optional, List, Dict

from langchain_community.chat_models import ChatOllama

from llama_index.llms.ollama import Ollama
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.embeddings.ollama import OllamaEmbedding

from pygpt_net.core.types import (
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
)
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem
import nest_asyncio


class OllamaLLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(OllamaLLM, self).__init__(*args, **kwargs)
        self.id = "ollama"
        self.type = [MODE_LANGCHAIN, MODE_LLAMA_INDEX, "embeddings"]

    def completion(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ):
        """
        Return LLM provider instance for completion

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        return None

    def chat(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ):
        """
        Return LLM provider instance for chat

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        args = self.parse_args(model.langchain)
        return ChatOllama(**args)

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        nest_asyncio.apply()
        args = self.parse_args(model.llama_index)
        if "request_timeout" not in args:
            args["request_timeout"] = 120
        return Ollama(**args)

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: Embedding provider instance
        """
        args = {}
        if config is not None:
            args = self.parse_args({
                "args": config,
            })
        return OllamaEmbedding(**args)

    def init_embeddings(
            self,
            window,
            env: Optional[List[Dict]] = None
    ):
        """
        Initialize embeddings provider

        :param window: window instance
        :param env: ENV configuration list
        """
        super(OllamaLLM, self).init_embeddings(window, env)

        # === FIX FOR LOCAL EMBEDDINGS ===
        # if there is no OpenAI api key then set fake key to prevent empty key Llama-index error
        if ('OPENAI_API_KEY' not in os.environ
                and (window.core.config.get('api_key') is None or window.core.config.get('api_key') == "")):
            os.environ['OPENAI_API_KEY'] = "_"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.01 11:00:00                  #
# ================================================== #

from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

from .base import Base
from .custom import Custom
from .template import Template


class Prompt:
    def __init__(self, window=None):
        """
        Prompt core

        :param window: Window instance
        """
        self.window = window
        self.base = Base(window)
        self.custom = Custom(window)
        self.template = Template(window)

    def get(
            self,
            prompt: str,
            mode: str = None,
            model: ModelItem = None
    ) -> str:
        """
        Get system prompt content

        :param prompt: id of the prompt
        :param mode: mode
        :param model: model item
        :return: text content
        """
        return self.base.get(
            prompt=prompt,
            mode=mode,
            model=model,
        )

    def build_final_system_prompt(
            self,
            prompt: str,
            mode: str,
            model: ModelItem = None
    ) -> str:
        """
        Build final system prompt

        :param prompt: prompt
        :param mode: mode
        :param model: model item
        :return: final system prompt
        """
        # tmp dispatch event: system prompt
        event = Event(Event.SYSTEM_PROMPT, {
            'mode': self.window.core.config.get('mode'),
            'value': prompt,
            'silent': True,
        })
        self.window.dispatch(event)
        prompt = event.data['value']

        if (self.window.core.config.get('cmd')
                or self.window.controller.plugins.is_type_enabled("cmd.inline")):

            # abort if native func call enabled
            if self.window.core.command.is_native_enabled():
                return prompt

            # abort if model not supported
            if not self.window.core.command.is_model_supports_tools(mode, model):
                return prompt

            # cmd syntax tokens
            data = {
                'prompt': prompt,
                'silent': True,
                'syntax': [],
                'cmd': [],
            }

            # IMPORTANT: append command syntax only if at least one command is detected
            # tmp dispatch event: command syntax apply
            # full execute cmd syntax
            if self.window.core.config.get('cmd'):
                event = Event(Event.CMD_SYNTAX, data)
                self.window.dispatch(event)
                if event.data and "cmd" in event.data and event.data["cmd"]:
                    prompt = self.window.core.command.append_syntax(
                        data=event.data,
                        mode=mode,
                        model=model,
                    )

            # inline cmd syntax only
            elif self.window.controller.plugins.is_type_enabled("cmd.inline"):
                event = Event(Event.CMD_SYNTAX_INLINE, data)
                self.window.dispatch(event)
                if event.data and "cmd" in event.data and event.data["cmd"]:
                    prompt = self.window.core.command.append_syntax(
                        data=event.data,
                        mode=mode,
                        model=model,
                    )

        return prompt

    def prepare_sys_prompt(
            self,
            mode: str,
            model: ModelItem,
            sys_prompt: str,
            ctx: CtxItem,
            reply: bool,
            internal: bool,
            is_expert: bool = False,
            disable_native_tool_calls: bool = False,
    ) -> str:
        """
        Prepare system prompt

        :param mode: mode
        :param model: model item
        :param sys_prompt: system prompt
        :param ctx: context item
        :param reply: reply from plugins
        :param internal: internal call
        :param is_expert: called from expert
        :param disable_native_tool_calls: True to disable native func calls
        :return: system prompt
        """
        # event: system prompt (append to system prompt)
        event = Event(Event.SYSTEM_PROMPT, {
            'mode': mode,
            'value': sys_prompt,
            'is_expert': is_expert,
        })
        self.window.dispatch(event)
        sys_prompt = event.data['value']

        # event: post prompt (post-handle system prompt)
        event = Event(Event.POST_PROMPT, {
            'mode': mode,
            'reply': reply,
            'internal': internal,
            'value': sys_prompt,
            'is_expert': is_expert,
        })
        event.ctx = ctx
        self.window.dispatch(event)
        sys_prompt = event.data['value']

        # event: command syntax apply (if commands enabled or inline plugin then append commands prompt)
        if self.window.core.config.get('cmd') or self.window.controller.plugins.is_type_enabled("cmd.inline"):
            if self.window.core.command.is_native_enabled() and not disable_native_tool_calls:
                return sys_prompt  # abort if native func call enabled

            # abort if model not supported
            if not self.window.core.command.is_model_supports_tools(mode, model):
                return sys_prompt

            data = {
                'mode': mode,
                'prompt': sys_prompt,
                'syntax': [],
                'cmd': [],
                'is_expert': is_expert,
            }
            # IMPORTANT: append command syntax only if at least one command is detected
            # full execute cmd syntax
            if self.window.core.config.get('cmd'):
                event = Event(Event.CMD_SYNTAX, data)
                self.window.dispatch(event)
                if event.data and "cmd" in event.data and event.data["cmd"]:
                    sys_prompt = self.window.core.command.append_syntax(
                        data=event.data,
                        mode=mode,
                        model=model,
                    )

            # inline cmd syntax only
            elif self.window.controller.plugins.is_type_enabled("cmd.inline"):
                event = Event(Event.CMD_SYNTAX_INLINE, data)
                self.window.dispatch(event)
                if event.data and "cmd" in event.data and event.data["cmd"]:
                    sys_prompt = self.window.core.command.append_syntax(
                        data=event.data,
                        mode=mode,
                        model=model,
                    )

        return sys_prompt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 04:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .runner import Runner
from .worker import Worker

from pygpt_net.utils import trans


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "mailer"
        self.name = "Mailer"
        self.description = "Provides email sending"
        self.prefix = "Email"
        self.type = [
            'email',
        ]
        self.order = 100
        self.allowed_cmds = [
            "send_mail",
            "get_emails",
            "get_email_body",
        ]
        self.use_locale = True
        self.runner = Runner(self)
        self.worker = None
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx
        silent = data.get("silent", False)

        if name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
                silent,
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for item in self.allowed_cmds:
            if self.has_cmd(item):
                cmd = self.get_cmd(item)
                data['cmd'].append(cmd)  # append command

    def cmd(self, ctx: CtxItem, cmds: list, silent: bool = False):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        :param silent: silent mode
        """
        is_cmd = False
        force = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True
                if "force" in item and item["force"]:
                    force = True  # call from tool

        if not is_cmd:
            return

        # set state: busy
        if not silent:
            self.cmd_prepare(ctx, my_commands)

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.cmds = my_commands
            worker.ctx = ctx

            # connect signals
            self.runner.attach_signals(worker.signals)

            if not self.is_async(ctx) and not force:
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import json
from typing import Optional, Dict, Any, List

from pygpt_net.core.events import KernelEvent, RenderEvent
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.item.ctx import CtxItem

class Reply:
    def __init__(self, window=None):
        """
        Reply handler

        :param window: Window instance
        """
        self.window = window
        self.reply_stack = []
        self.reply_ctx = None
        self.last_result = None
        self.reply_idx = -1

    def add(
            self,
            context: BridgeContext,
            extra: Dict[str, Any] = None
    ) -> List[Dict]:
        """
        Send reply from plugins to model

        :param context: bridge context
        :param extra: extra data
        :return: list of results
        """
        flush = False
        if "flush" in extra and extra["flush"]:
            flush = True
        ctx = context.ctx
        if ctx is not None:
            self.run_post_response(ctx, extra)
            self.last_result = ctx.results
            if ctx.agent_call:
                return ctx.results # abort if called in agent and return here, TODO: check if needed!!!!!
            self.window.core.debug.info("Reply...")
            if self.window.core.debug.enabled() and self.is_log():
                self.window.core.debug.debug("CTX REPLY: " + str(ctx))
            if ctx.reply:
                if self.reply_idx >= ctx.pid:  # skip if reply already sent for this context
                    # >>> this prevents multiple replies from the same ctx item <<<
                    return []
                self.reply_idx = ctx.pid 
                self.append(ctx)
            if flush or self.window.controller.kernel.async_allowed(ctx):
                self.flush()
            return ctx.results
        return []

    def append(self, ctx: CtxItem):
        """
        Add reply to stack

        :param ctx: context item
        """
        self.window.core.debug.info("Reply stack (add)...")
        self.reply_stack.append(ctx.results)
        ctx.results = []  # clear results
        self.reply_ctx = ctx

    def flush(self):
        """Flush reply stack"""
        if self.reply_ctx is None or len(self.reply_stack) == 0:
            return

        self.window.core.debug.info("Reply stack (flush)...")
        results = []
        for responses in self.reply_stack:
            for result in responses:
                results.append(result)

        self.window.update_status("")  # clear status
        if self.reply_ctx.internal:
            if self.window.controller.agent.legacy.enabled():
                self.window.controller.agent.legacy.add_run()
                self.window.controller.agent.legacy.update()

        # prepare data to send as reply
        tool_data = json.dumps(results)
        if (len(self.reply_stack) < 2
                and self.reply_ctx.extra_ctx
                and self.window.core.config.get("ctx.use_extra")):
            tool_data = self.reply_ctx.extra_ctx  # if extra content is set, use it as data to send

        prev_ctx = self.window.core.ctx.as_previous(self.reply_ctx)  # copy result to previous ctx and clear current ctx
        self.window.core.ctx.update_item(self.reply_ctx)  # update context in db
        self.window.update_status('...')

        # if response from sub call, from experts
        parent_id = None
        if self.reply_ctx.sub_call:
            if self.reply_ctx.meta is not None:
                parent_id = self.reply_ctx.meta.id  # slave meta id

        # tool output append
        data = {
            "meta": self.reply_ctx.meta,
            "tool_data": tool_data,
        }
        event = RenderEvent(RenderEvent.TOOL_UPDATE, data)
        self.window.dispatch(event)
        self.clear()

        # send reply
        context = BridgeContext()
        context.ctx = prev_ctx
        context.prompt = str(tool_data)
        extra = {
            "force": True,
            "reply": True,
            "internal": True,
            "parent_id": parent_id,
        }
        event = KernelEvent(KernelEvent.REPLY_RETURN, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)

    def run_post_response(
            self,
            ctx: CtxItem,
            extra_data: Optional[Dict[str, Any]] = None
    ):
        """
        Run post-response operations

        :param ctx: context (CtxItem)
        :param extra_data: extra data
        """
        if isinstance(extra_data, dict):
            if (ctx is None or not ctx.agent_call) or not self.window.controller.kernel.is_threaded():
                if "post_update" in extra_data and isinstance(extra_data["post_update"], list):
                    if "file_explorer" in extra_data["post_update"]:
                        # update file explorer view
                        self.window.controller.files.update_explorer()

    def clear(self):
        """Clear reply stack"""
        self.window.core.debug.info("Reply stack (clear)...")
        self.reply_ctx = None
        self.reply_stack = []

    def is_log(self) -> bool:
        """
        Check if event can be logged

        :return: true if can be logged
        """
        is_log = False
        if self.window.core.config.has("log.events") \
                and self.window.core.config.get("log.events"):
            is_log = True
        return is_log



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.05 23:00:00                  #
# ================================================== #

from typing import Optional, Tuple

from PySide6.QtGui import QTextCursor

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.item.notepad import NotepadItem
from pygpt_net.ui.widget.tabs.body import TabBody
from pygpt_net.ui.widget.textarea.notepad import NotepadWidget
from pygpt_net.utils import trans

import pygpt_net.icons_rc


class Notepad:
    def __init__(self, window=None):
        """
        Notepad controller

        :param window: Window instance
        """
        self.window = window
        self.opened_once = False
        self.opened_idx = []

    def create(
            self,
            idx: Optional[int] = None
    ) -> Tuple[TabBody, int, int]:
        """
        Create notepad widget

        :param idx: notepad idx
        :return: notepad widget (TabBody)
        """
        if idx is None:
            idx = self.window.core.tabs.count_by_type(Tab.TAB_NOTEPAD) + 1
        data_id = idx
        self.window.ui.notepad[data_id] = NotepadWidget(self.window)
        self.window.ui.notepad[data_id].id = idx
        self.window.ui.notepad[data_id].textarea.id = idx
        title = trans('output.tab.notepad')
        title += " " + str(idx)
        children = self.window.core.tabs.from_widget(self.window.ui.notepad[data_id])
        return children, idx, data_id

    def load(self):
        """Load all notepads contents"""
        self.window.core.notepad.load_all()
        items = self.window.core.notepad.get_all()
        num_notepads = self.get_num_notepads()
        if len(items) == 0:
            if num_notepads > 0:
                for idx in range(1, num_notepads + 1):
                    item = NotepadItem()
                    item.idx = idx
                    items[idx] = item

        if num_notepads > 0:
            for idx in range(1, num_notepads + 1):
                if idx not in items:
                    item = NotepadItem()
                    item.idx = idx
                    items[idx] = item
                if idx in self.window.ui.notepad:
                    self.window.ui.notepad[idx].setText(items[idx].content)

    def get_notepad_name(self, idx: int):
        """
        Get notepad name

        :param idx: notepad idx
        :return: notepad name
        """
        num = self.get_num_notepads()
        if num > 1:
            title = trans('text.context_menu.copy_to.notepad') + ' ' + str(idx)
        else:
            title = trans('text.context_menu.copy_to.notepad')
        item = self.window.core.notepad.get_by_id(idx)
        if item is None:
            return None
        if item.initialized and item.title is not None and len(item.title) > 0:
            title = item.title
        return title

    def save(self, idx: int):
        """
        Save notepad contents

        :param idx: notepad idx
        """
        item = self.window.core.notepad.get_by_id(idx)
        if item is None:
            item = NotepadItem()
            item.idx = idx
            self.window.core.notepad.items[idx] = item

        if idx in self.window.ui.notepad:
            prev_content = item.content
            item.content = self.window.ui.notepad[idx].toPlainText()
            if prev_content != item.content:  # update only if content changed
                self.window.core.notepad.update(item)
            self.update()

    def save_all(self):
        """Save all notepads contents"""
        items = self.window.core.notepad.get_all()
        num_notepads = self.get_num_notepads()
        if num_notepads > 0:
            tabs = self.window.core.tabs.get_tabs_by_type(Tab.TAB_NOTEPAD)
            for tab in tabs:
                idx = tab.data_id
                if idx in self.window.ui.notepad:
                    prev_content = str(items[idx].content)
                    items[idx].content = self.window.ui.notepad[idx].toPlainText()
                    # update only if content changed
                    if prev_content != items[idx].content:
                        self.window.core.notepad.update(items[idx])
            self.update()

    def setup(self):
        """Setup all notepads"""
        self.load()

    def append_text(self, text: str, idx: int):
        """
        Append text to notepad

        :param text: text to append
        :param idx: notepad idx
        """
        if idx not in self.window.ui.notepad:
            return
        dt = ""  # TODO: add to config append date/time
        # dt = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + ":\n--------------------------\n"
        prev_text = self.window.ui.notepad[idx].toPlainText()
        if prev_text.strip() != "":
            prev_text += "\n"
        new_text = prev_text + dt + text.strip()
        self.window.ui.notepad[idx].setText(new_text)
        self.save(idx)

        # move cursor to end
        cursor = self.window.ui.notepad[idx].textarea.textCursor()
        cursor.movePosition(QTextCursor.End)
        self.window.ui.notepad[idx].textarea.setTextCursor(cursor)

    def get_num_notepads(self) -> int:
        """
        Get number of notepads

        :return: number of notepads
        """
        return self.window.core.tabs.count_by_type(Tab.TAB_NOTEPAD)

    def get_current_active(self) -> int:
        """
        Get current notepad idx

        :return: current notepad index
        """
        if self.is_active():
            tab = self.window.controller.ui.tabs.get_current_tab()
            if tab is not None:
                return tab.data_id
        return 1

    def is_active(self) -> bool:
        """
        Check if notepad tab is active

        :return: True if notepad tab is active
        """
        return self.window.controller.ui.tabs.get_current_type() == Tab.TAB_NOTEPAD

    def open(self):
        """Open notepad"""
        # if notepad disabled, do nothing
        if self.get_num_notepads() == 0:
            return

        # switch to first notepad tab if current tab is not notepad
        tabs = self.window.ui.layout.get_active_tabs()
        if self.window.controller.ui.tabs.get_current_type() != Tab.TAB_NOTEPAD:
            idx = self.window.core.tabs.get_min_idx_by_type(Tab.TAB_NOTEPAD)
            if idx is not None:
                tabs.setCurrentIndex(idx)

        self.window.activateWindow()  # focus

    def update(self):
        """Update notepads UI"""
        pass

    def reload(self):
        """Reload notepads"""
        self.window.core.notepad.locked = True
        self.window.core.notepad.reset()
        self.load()
        self.window.core.notepad.locked = False

    def switch_to_tab(self, idx: Optional[int] = None):
        """
        Switch to notepad tab

        :param idx: notepad idx
        """
        if idx is None:
            idx = self.get_first_notepad_tab_idx()
        tabs = self.window.ui.layout.get_active_tabs()
        tab = self.window.core.tabs.get_tab_by_index(idx)
        if tab is not None:
            tabs.setCurrentIndex(idx)
        else:
            tabs.setCurrentIndex(self.get_first_notepad_tab_idx())

    def get_first_notepad_tab_idx(self) -> int:
        """
        Get first notepad tab index

        :return: first notepad tab index
        """
        return self.window.core.tabs.get_min_idx_by_type(Tab.TAB_NOTEPAD)

    def get_current_notepad_text(self) -> str:
        """
        Get current notepad text

        :return: current notepad text
        """
        idx = self.get_current_active()
        if idx in self.window.ui.notepad:
            return self.window.ui.notepad[idx].toPlainText()
        return ""

    def get_notepad_text(self, idx: int) -> str:
        """
        Get notepad text

        :param idx: notepad index
        :return: notepad text
        """
        if idx in self.window.ui.notepad:
            return self.window.ui.notepad[idx].toPlainText()
        return ""

    def clear(self, idx: int) -> bool:
        """
        Clear notepad contents

        :param idx: notepad idx
        """
        if idx in self.window.ui.notepad:
            self.window.ui.notepad[idx].textarea.clear()
            self.save(idx)
            return True
        return False

    def on_open(
            self,
            tab_idx: int,
            column_idx: int = 0
    ):
        """
        On open notepad tab

        :param tab_idx: current tab idx
        :param column_idx: column idx
        """
        tab = self.window.controller.ui.tabs.get_current_tab()
        if tab is None:
            return
        if tab.type == Tab.TAB_NOTEPAD:
            idx = tab.data_id
            if idx in self.window.ui.notepad:
                if idx not in self.opened_idx:
                    self.window.ui.notepad[idx].scroll_to_bottom()
                if not self.window.ui.notepad[idx].opened:
                    self.window.ui.notepad[idx].opened = True
                if idx not in self.opened_idx:
                    self.opened_idx.append(idx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from datetime import datetime
from typing import Optional, List

from PySide6.QtGui import QTextCursor, QTextBlockFormat

from pygpt_net.core.render.base import BaseRenderer
from pygpt_net.ui.widget.textarea.input import ChatInput
from pygpt_net.ui.widget.textarea.output import ChatOutput
from pygpt_net.item.ctx import CtxItem, CtxMeta

from .body import Body
from .helpers import Helpers
from .pid import PidData


class Renderer(BaseRenderer):
    def __init__(self, window=None):
        super(Renderer, self).__init__(window)
        """
        Plain text renderer

        :param window: Window instance
        """
        self.window = window
        self.body = Body(window)
        self.helpers = Helpers(window)
        self.pids = {}  # per node data

    def prepare(self):
        """
        Prepare renderer
        """
        self.pids = {}

    def get_pid(self, meta: CtxMeta):
        """
        Get PID for context meta
        
        :param meta: context PID
        """
        return self.window.core.ctx.output.get_pid(meta)

    def get_or_create_pid(self, meta: CtxMeta):
        """
        Get PID for context meta and create PID data (if not exists)
        
        :param meta: context PID
        """
        if meta is not None:
            pid = self.get_pid(meta)
            if pid not in self.pids:
                self.pid_create(pid, meta)
            return pid

    def pid_create(self, pid, meta: CtxMeta):
        """
        Create PID data
        
        :param pid: PID
        :param meta: context meta
        """
        if pid is not None:
            self.pids[pid] = PidData(pid, meta)

    def begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render begin
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.to_end(meta)

    def end(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end
            
        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.to_end(meta)

    def end_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end extra
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.to_end(meta)

    def stream_begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream begin

        :param meta: context meta
        :param ctx: context item
        """
        pass  # do nothing

    def stream_end(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream end

        :param meta: context meta
        :param ctx: context item
        """
        pass  # do nothing

    def append_context(
            self,
            meta: CtxMeta,
            items: List[CtxItem],
            clear: bool = True
    ):
        """
        Append all context to output

        :param meta: context meta
        :param items: context items
        :param clear: True if clear all output before append
        """
        if clear:
            self.clear_output(meta)

        i = 0
        for item in items:
            item.idx = i
            self.append_context_item(meta, item)
            i += 1

    def append_input(
            self,
            meta: CtxMeta,
            item: CtxItem,
            flush: bool = True,
            append: bool = False
    ):
        """
        Append text input to output

        :param meta: context meta
        :param item: context item
        :param flush: True if flush
        :param append: True to force append node
        """
        if item.input is None or item.input == "":
            return
        if self.is_timestamp_enabled() and item.input_timestamp is not None:
            name = ""
            if item.input_name is not None and item.input_name != "":
                name = item.input_name + " "
            ts = datetime.fromtimestamp(item.input_timestamp)
            hour = ts.strftime("%H:%M:%S")
            text = '{}{} > {}'.format(name, hour, item.input)
        else:
            text = "> {}".format(item.input)
        self.append_raw(meta, item, text.strip())
        self.to_end(meta)

    def append_output(
            self,
            meta: CtxMeta,
            item: CtxItem
    ):
        """
        Append text output to output

        :param meta: context meta
        :param item: context item
        """
        if item.output is None or item.output == "":
            return
        if self.is_timestamp_enabled() and item.output_timestamp is not None:
            name = ""
            if item.output_name is not None and item.output_name != "":
                name = item.output_name + " "
            ts = datetime.fromtimestamp(item.output_timestamp)
            hour = ts.strftime("%H:%M:%S")
            text = '{}{} {}'.format(name, hour, item.output)
        else:
            text = "{}".format(item.output)
        self.append_raw(meta, item, text.strip())
        self.to_end(meta)

    def append_extra(
            self,
            meta: CtxMeta,
            item: CtxItem,
            footer: bool = False
    ):
        """
        Append extra data (images, files, etc.) to output

        :param meta: context meta
        :param item: context item
        :param footer: True if it is a footer
        """
        appended = []
        pid = self.get_or_create_pid(meta)

        # images
        c = len(item.images)
        if c > 0:
            n = 1
            for image in item.images:
                if image in appended or image in self.pids[pid].images_appended:
                    continue
                try:
                    appended.append(image)
                    self.append_raw(meta, item, self.body.get_image_html(image, n, c))
                    self.pids[pid].images_appended.append(image)
                    n += 1
                except Exception as e:
                    pass

        # files and attachments, TODO check attachments
        c = len(item.files)
        if c > 0:
            n = 1
            for file in item.files:
                if file in appended:
                    continue
                try:
                    appended.append(file)
                    self.append_raw(meta, item, self.body.get_file_html(file, n, c))
                    n += 1
                except Exception as e:
                    pass

        # urls
        c = len(item.urls)
        if c > 0:
            urls_str = []
            n = 1
            for url in item.urls:
                if url in appended or url in self.pids[pid].urls_appended:
                    continue
                try:
                    appended.append(url)
                    urls_str.append(self.body.get_url_html(url, n, c))
                    self.pids[pid].urls_appended.append(url)
                    n += 1
                except Exception as e:
                    pass
            if urls_str:
                self.append_raw(meta, item, "\n" + "\n".join(urls_str))

        # docs json
        if self.window.core.config.get('ctx.sources'):
            if item.doc_ids is not None and len(item.doc_ids) > 0:
                try:
                    docs = self.body.get_docs_html(item.doc_ids)
                    self.get_output_node(meta).append(docs)
                    self.to_end(meta)
                except Exception as e:
                    pass

        # jump to end
        if len(appended) > 0:
            self.to_end(meta)

    def append_chunk(
            self,
            meta: CtxMeta,
            item: CtxItem,
            text_chunk: str,
            begin: bool = False
    ):
        """
        Append output chunk to output

        :param meta: context meta
        :param item: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        if text_chunk is None or text_chunk == "":
            return

        pid = self.get_or_create_pid(meta)
        raw_chunk = str(text_chunk)

        if begin:
            self.pids[pid].buffer = ""  # reset buffer
            self.pids[pid].is_cmd = False  # reset command flag

            if self.is_timestamp_enabled() and item.output_timestamp is not None:
                name = ""
                if item.output_name is not None and item.output_name != "":
                    name = item.output_name + " "
                ts = datetime.fromtimestamp(item.output_timestamp)
                hour = ts.strftime("%H:%M:%S")
                text_chunk = "{}{}: ".format(name, hour) + text_chunk

            text_chunk = "\n" + text_chunk
            self.append_block(meta)
            self.append_chunk_start(meta, item)

        self.pids[pid].buffer += raw_chunk
        self.append(meta, item, self.helpers.format_chunk(text_chunk), "")

    def append_block(
            self,
            meta: CtxMeta
    ):
        """
        Append block to output

        :param meta: context meta
        :param ctx: context item
        """
        node = self.get_output_node(meta)
        cursor = node.textCursor()
        cursor.movePosition(QTextCursor.End)
        block_format = QTextBlockFormat()
        block_format.setIndent(0)
        cursor.insertBlock(block_format)
        node.setTextCursor(cursor)

    def append_raw(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            text: str
    ):
        """
        Append and format raw text to output
        
        :param meta: context meta
        :param ctx: context item
        :param text: text to append
        """
        node = self.get_output_node(meta)
        prev_text = node.toPlainText()
        if prev_text != "":
            prev_text += "\n\n"
        new_text = prev_text + text.strip()
        node.setText(new_text)
        cur = node.textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)

    def append_chunk_start(self, meta: CtxMeta, ctx: CtxItem):
        """
        Append start of chunk to output
        
        :param meta: context meta
        :param ctx: context item
        """
        node = self.get_output_node(meta)
        cursor = node.textCursor()
        cursor.movePosition(QTextCursor.End)
        node.setTextCursor(cursor)

    def append_context_item(
            self,
            meta: CtxMeta,
            item: CtxItem
    ):
        """
        Append context item to output
        
        :param meta: context meta
        :param item: context item
        """
        self.append_input(meta, item)
        self.append_output(meta, item)
        self.append_extra(meta, item)

    def append(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            text: str,
            end: str = "\n"
    ):
        """
        Append text to output

        :param meta: context meta
        :param ctx: context item
        :param text: text to append
        :param end: end of the line character
        """
        node = self.get_output_node(meta)
        cur = node.textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)
        s = str(text) + end
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertText("\n")
        node.setTextCursor(cur)  # Update visible cursor

    def append_timestamp(
            self,
            item: CtxItem,
            text: str
    ) -> str:
        """
        Append timestamp to text
        
        :param item: context item
        :param text: input text
        :return: Text with timestamp (if enabled)
        """
        if item is not None \
                and self.is_timestamp_enabled() \
                and item.input_timestamp is not None:
            ts = datetime.fromtimestamp(item.input_timestamp)
            hour = ts.strftime("%H:%M:%S")
            text = '{}: {}'.format(hour, text)
        return text

    def reset(self, meta: Optional[CtxMeta] = None):
        """
        Reset

        :param meta: context meta
        """
        pid = self.get_or_create_pid(meta)
        if pid is not None:
            self.pids[pid].images_appended = []
            self.pids[pid].urls_appended = []

    def reload(self):
        """Reload output, called externally only on theme change to redraw content"""
        self.window.controller.ctx.refresh_output()  # if clear all and appends all items again

    def clear_output(self, meta: Optional[CtxMeta] = None):
        """
        Clear output

        :param meta: context meta
        """
        self.reset()
        self.get_output_node(meta).clear()

    def clear_input(self):
        """Clear input"""
        self.get_input_node().clear()

    def to_end(self, meta: CtxMeta):
        """
        Move cursor to end of output

        :param meta: context meta
        """
        node = self.get_output_node(meta)
        cursor = node.textCursor()
        cursor.movePosition(QTextCursor.End)
        node.setTextCursor(cursor)

    def is_timestamp_enabled(self) -> bool:
        """
        Check if timestamp is enabled

        :return: True if timestamp is enabled
        """
        return self.window.core.config.get('output_timestamp')

    def get_output_node(
            self,
            meta: Optional[CtxMeta] = None
    ) -> ChatOutput:
        """
        Get output node for current context
        
        :param meta: context meta
        :return: output node
        """
        return self.window.core.ctx.output.get_current_plain(meta)

    def get_input_node(self) -> ChatInput:
        """
        Get input node

        :return: input node
        """
        return self.window.ui.nodes['input']

    def get_all_nodes(self) -> list:
        """
        Return all registered nodes

        :return: list of ChatOutput nodes (tabs)
        """
        return self.window.core.ctx.output.get_all_plain()

    def clear_all(self):
        """Clear all"""
        for node in self.get_all_nodes():
            node.clear()
            node.document().setDefaultStyleSheet("")
            node.setStyleSheet("")
            node.document().setMarkdown("")
            node.document().setHtml("")
            node.setPlainText("")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import copy
import os
import shutil
import uuid

from shutil import copyfile
from typing import Optional, List, Dict, Any, Tuple

from llama_index.core import Document

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.events import KernelEvent
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxMeta, CtxItem


class Context:
    def __init__(self, window=None):
        """
        Context attachment core

        :param window: Window instance
        """
        self.window = window
        self.dir_index = "index"
        self.last_used_item = None
        self.last_used_content = None
        self.last_used_context = None
        self.last_files = []
        self.last_urls = []
        self.summary_prompt = """
        Summarize the text below by extracting the most important information, 
        especially those that may help answer the question: 

        `{query}`

        If the answer to the question is not in the text to summarize, 
        simply return a summary of the entire content. 
        Do not include anything other than the summary itself in your response. 
        The response should not contain any phrases other than the summary of the content itself:        

        # Example of a bad response:        

        `According to the text you sent, item X is located at place Y.`        

        # Example of a correct response:        

        `[filename] Item X is located at place Y.`        

        # Content to summarize:

        `{content}`
        """
        self.rag_prompt = """
        Prepare a question for the RAG engine (vector database) asking for additional context that can help obtain 
        extra information necessary to answer the user's question. The query should be brief and to the point, 
        so as to be processed as effectively as possible by the RAG engine. Below is the entire conversation 
        of the user with the AI assistant, and at the end the current user's question, for which you need to 
        prepare DIRECT query for the RAG engine for additional context, taking into account the content of the entire 
        discussion and its context. In your response, return only the DIRECT query for additional context, 
        do not return anything else besides it. The response should not contain any phrases other than the query itself:
        
        # Good RAG query example:
        
        `What is the capital of France?`
        
        # Bad RAG query example:
        
        `Can you tell me the capital of France?`

        # Full conversation:
        
        `{history}`
        
        # User question:
        
        `{query}`
        """

    def get_context(
            self,
            mode: str,
            ctx: CtxItem,
            history: List[CtxItem]
    ) -> str:
        """
        Get context for mode

        :param mode: Context mode
        :param ctx: CtxItem instance
        :param history: history
        :return: context
        """
        content = ""
        if mode == self.window.controller.chat.attachment.MODE_FULL_CONTEXT:
            content = self.get_context_text(ctx, filename=True)
        elif mode == self.window.controller.chat.attachment.MODE_QUERY_CONTEXT:
            content = self.query_context(ctx, history)
        elif mode == self.window.controller.chat.attachment.MODE_QUERY_CONTEXT_SUMMARY:
            content = self.summary_context(ctx, history)
        return content

    def get_context_text(
            self,
            ctx: CtxItem,
            filename: bool = False
    ) -> str:
        """
        Get raw text context for meta

        :param ctx: CtxItem instance
        :param filename: append filename
        :return: raw context
        """
        meta = ctx.meta
        meta_path = self.get_dir(meta)
        context = ""
        if os.path.exists(meta_path) and os.path.isdir(meta_path):
            for file in meta.get_additional_ctx():
                if ("type" not in file
                        or file["type"] not in ["local_file", "url"]):
                    continue
                if not "uuid" in file:
                    continue
                file_id = file["uuid"]
                file_idx_path = os.path.join(meta_path, file_id)
                text_path = os.path.join(file_idx_path, file_id + ".txt")
                store_path = file["path"]
                if "real_path" in file:
                    store_path = file["real_path"]
                if filename:
                    if file["type"] == "url":
                        context += "URL: {}\n".format(file["path"]) + "\n"
                    else:
                        context += "Filename: {}\n".format(file["name"]) + "\n"

                # store used files and URLs in ctx
                if file["type"] == "url":
                    if store_path not in self.last_urls:
                        self.last_urls.append(store_path)
                else:
                    if store_path not in self.last_files:
                        self.last_files.append(store_path)

                if os.path.exists(text_path):
                    try:
                        with open(text_path, "r", encoding="utf-8") as f:
                            context += f.read() + "\n\n"
                    except Exception as e:
                        print("Attachments: read error: {}".format(e))

        self.last_used_content = context
        self.last_used_context = context
        return context

    def query_context(
            self,
            ctx: CtxItem,
            history: List[CtxItem]
    ) -> str:
        """
        Query the index for context

        :param ctx: CtxItem instance
        :param history: history
        :return: query result
        """
        meta = ctx.meta
        meta_path = self.get_dir(meta)
        query = str(ctx.input)
        if not os.path.exists(meta_path) or not os.path.isdir(meta_path):
            return ""
        idx_path = os.path.join(self.get_dir(meta), self.dir_index)

        indexed = False
        # index files if not indexed by auto_index
        for i, file in enumerate(meta.get_additional_ctx()):
            if "indexed" not in file or not file["indexed"]:
                file_id = file["uuid"]
                file_idx_path = os.path.join(meta_path, file_id)
                file_path = os.path.join(file_idx_path, file["name"])
                type = AttachmentItem.TYPE_FILE
                source = file_path
                if "type" in file:
                    if file["type"] == "url":
                        type = AttachmentItem.TYPE_URL
                        source = file["path"] # URL
                doc_ids = self.index_attachment(type, source, idx_path)
                file["indexed"] = True
                file["doc_ids"] = doc_ids
                indexed = True

        if indexed:
            # update ctx in DB
            self.window.core.ctx.replace(meta)
            self.window.core.ctx.save(meta.id)

        history_data = self.prepare_context_history(history)
        model, model_item = self.get_selected_model("query")

        verbose = False
        if self.is_verbose():
            verbose = True
            print("Attachments: using query model: {}".format(model))

        result = self.window.core.idx.chat.query_attachment(
            query=query,
            path=idx_path,
            model=model_item,
            history=history_data,
            verbose=verbose,
        )
        self.last_used_context = result

        if self.is_verbose():
            print("Attachments: query result: {}".format(result))

        return result

    def summary_context(
            self,
            ctx: CtxItem,
            history: List[CtxItem]
    ) -> str:
        """
        Get summary of the context

        :param ctx: CtxItem instance
        :param history: history
        :return: query result
        """
        model, model_item = self.get_selected_model("summary")
        if model_item is None:
            raise Exception("Attachments: summary model not found: {}".format(model))

        if self.is_verbose():
            print("Attachments: using summary model: {}".format(model))

        query = str(ctx.input)
        content = self.get_context_text(ctx, filename=True)
        prompt = self.summary_prompt.format(
            query=str(query).strip(),
            content=str(content).strip(),
        )
        if self.is_verbose():
            print("Attachments: summary prompt: {}".format(prompt))

        history_data = self.prepare_context_history(history)
        ctx = CtxItem()
        bridge_context = BridgeContext(
            ctx=ctx,
            prompt=prompt,
            stream=False,
            model=model_item,
            history=history_data,
        )
        event = KernelEvent(KernelEvent.CALL, {
            'context': bridge_context,
            'extra': {},
        })
        self.window.dispatch(event)
        response = event.data.get("response")
        self.last_used_context = response
        if self.is_verbose():
            print("Attachments: summary received: {}".format(response))
        return response

    def prepare_context_history(
            self,
            history: List[CtxItem]
    ) -> List[CtxItem]:
        """
        Prepare context history

        :param history: history
        :return: history data
        """
        use_history = self.window.core.config.get("ctx.attachment.rag.history", True)
        history_data = []
        if use_history:
            if self.is_verbose():
                print("Attachments: using history for query prepare...")

            # use only last X items from history
            num_items = self.window.core.config.get("ctx.attachment.rag.history.max_items", 3)
            history_data = []
            for item in history:
                history_data.append(item)

            # 0 = unlimited
            if num_items > 0:
                if self.is_verbose():
                    print("Attachments: using last {} items from history...".format(num_items))
                if len(history_data) < num_items:
                    num_items = len(history_data)
                history_data = history_data[-num_items:]

        return history_data

    def upload(
            self,
            meta: CtxMeta,
            attachment: AttachmentItem,
            prompt: str,
            auto_index: bool = False,
            real_path: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Upload attachment for context

        :param meta: CtxMeta instance
        :param attachment: AttachmentItem instance
        :param prompt: user input prompt
        :param auto_index: auto index
        :param real_path: real path
        :return: Dict with attachment data
        """
        if self.is_verbose():
            if meta.group:
                print("Uploading for meta group ID: {}".format(meta.group.id))
            else:
                print("Uploading for meta ID: {}".format(meta.id))

        # prepare idx dir
        name = os.path.basename(attachment.path)
        file_id = str(uuid.uuid4())
        meta_path = self.get_dir(meta)
        file_idx_path = os.path.join(meta_path, file_id)
        index_path = os.path.join(meta_path, self.dir_index)

        os.makedirs(meta_path, exist_ok=True)
        os.makedirs(file_idx_path, exist_ok=True)

        if self.is_verbose():
            print("Attachments: created path: {}".format(meta_path))
            if auto_index:
                print("Attachments: vector index path: {}".format(index_path))

        documents = None
        
        # store content to read, and get docs if type == web
        src_file, docs = self.store_content(attachment, file_idx_path)
        if attachment.type == AttachmentItem.TYPE_URL:
            documents = docs

        # extract text content using data loader, and get docs if type == file
        content, docs = self.read_content(attachment, src_file, prompt)
        if attachment.type == AttachmentItem.TYPE_FILE:
            documents = docs

        if content:
            text_path = os.path.join(file_idx_path, file_id + ".txt")
            with open(text_path, "w", encoding="utf-8") as f:
                f.write(content)
            if self.is_verbose():
                print("Attachments: read text content: {}".format(content))

        tokens = 0
        if content:
            tokens = self.window.core.tokens.from_str(content)

        type = "local_file"
        size = 0
        if attachment.type == AttachmentItem.TYPE_FILE:
            size = os.path.getsize(attachment.path)
        elif attachment.type == AttachmentItem.TYPE_URL:
            size = os.path.getsize(src_file)
            type = "url"  # extra ctx type

        # index file to ctx index
        doc_ids = []
        if auto_index:
            source = src_file
            if attachment.type == AttachmentItem.TYPE_URL:
                source = attachment.path  # URL
            doc_ids = self.index_attachment(attachment.type, source, index_path, documents=documents)

        result = {
            "name": name,
            "path": attachment.path,
            "type": type,
            "uuid": str(file_id),
            "content_type": "text",
            "size": size,
            "length": len(content),
            "tokens": tokens,
            "indexed": False,
        }
        if auto_index:
            result["indexed"] = True
            result["doc_ids"] = doc_ids

        if real_path:
            result["real_path"] = real_path

        if self.is_verbose():
            print("Attachments: uploaded: {}".format(result))

        return result

    def read_content(
            self,
            attachment: AttachmentItem,
            path: str,
            prompt: str
    ) -> Tuple[str, List[Document]]:
        """
        Read content from attachment

        :param attachment: AttachmentItem instance
        :param path: source file path
        :param prompt: user input prompt
        :return: text content, list of documents
        """
        content = ""
        docs = []
        if attachment.type == AttachmentItem.TYPE_FILE:
            loader_kwargs = {
                "prompt": prompt,
            }  # extra loader kwargs
            content, docs = self.window.core.idx.indexing.read_text_content(
                path=path,
                loader_kwargs=loader_kwargs,
            )
        elif attachment.type == AttachmentItem.TYPE_URL:
            # directly from before stored path
            with open(path, "r", encoding="utf-8") as f:
                content = f.read()  # data is already crawled in `store_content`
        return content, docs

    def store_content(
            self,
            attachment: AttachmentItem,
            dir: str
    ) -> Tuple[str, List[Document]]:
        """
        Prepare content for attachment

        :param attachment: AttachmentItem instance
        :param dir: directory to save content
        :return: path, list of documents
        """
        path = None
        docs = []
        if attachment.type == AttachmentItem.TYPE_FILE:
            # copy raw file only
            name = os.path.basename(attachment.path)
            path = os.path.join(dir, name)
            if os.path.exists(path):
                os.remove(path)
            copyfile(attachment.path, path)
        elif attachment.type == AttachmentItem.TYPE_URL:
            loader = attachment.extra.get("loader")
            input_params = attachment.extra.get("input_params")
            input_config = attachment.extra.get("input_config")
            self.window.core.idx.indexing.update_loader_args(loader, input_config)  # update config
            content, docs = self.window.core.idx.indexing.read_web_content(
                url="",
                type=loader,
                extra_args=input_params,
            )
            # src file save
            name = "url.txt"
            path = os.path.join(dir, name)
            if os.path.exists(path):
                os.remove(path)
            with open(path, "w", encoding="utf-8") as f:
                f.write(content)
        return path, docs

    def index_attachment(
            self,
            type: str,
            source: str,
            idx_path: str,
            documents: Optional[List[Document]] = None
    ) -> list:
        """
        Index attachment

        :param type: attachment type
        :param source: source file or URL
        :param idx_path: index path
        :param documents: list of documents (optional)
        :return: list of doc IDs
        """
        model = None
        doc_ids = []
        if type == AttachmentItem.TYPE_FILE:
            doc_ids = self.window.core.idx.indexing.index_attachment(source, idx_path, model, documents)
        elif type == AttachmentItem.TYPE_URL:
            doc_ids = self.window.core.idx.indexing.index_attachment_web(source, idx_path, model, documents)
        if self.is_verbose():
            print("Attachments: indexed. Doc IDs: {}".format(doc_ids))
        return doc_ids

    def get_all(self, meta: CtxMeta) -> list:
        """
        Get all attachments for meta

        :param meta: CtxMeta instance
        :return: list of attachments
        """
        return meta.get_additional_ctx()

    def get_dir(self, meta: CtxMeta) -> str:
        """
        Get directory for meta

        :param meta: CtxMeta instance
        :return: directory path
        """
        meta_uuid = str(meta.uuid)
        if meta.group:
            meta_uuid = str(meta.group.uuid)
        return os.path.join(self.window.core.config.get_user_dir("ctx_idx"), meta_uuid)

    def get_selected_model(self, mode: str = "summary"):
        """
        Get selected model for attachments

        :return: model name, model item
        """
        model_item = None
        model = None
        if mode == "summary":
            model = self.window.core.config.get("ctx.attachment.summary.model", "gpt-4o-mini")
        elif mode == "query":
            model = self.window.core.config.get("ctx.attachment.query.model", "gpt-4o-mini")
        if model:
            model_item = self.window.core.models.get(model)
        return model, model_item

    def duplicate(
            self,
            from_meta_id: int,
            to_meta_id: int
    ) -> bool:
        """
        Duplicate attachments from one meta to another

        :param from_meta_id: From meta id
        :param to_meta_id: To meta id
        :return
        """
        from_meta = self.window.core.ctx.get_meta_by_id(from_meta_id)
        to_meta = self.window.core.ctx.get_meta_by_id(to_meta_id)

        if from_meta is None or to_meta is None:
            return False
        for item in from_meta.additional_ctx:
            to_meta.additional_ctx.append(copy.deepcopy(item))
        self.window.core.ctx.save(to_meta.id)

        # copy index directory
        from_meta_path = self.get_dir(from_meta)
        to_meta_path = self.get_dir(to_meta)
        if os.path.exists(from_meta_path) and os.path.isdir(from_meta_path):
            shutil.copytree(from_meta_path, to_meta_path)
            if self.is_verbose():
                print("Attachments copied from {} to: {}".
                                           format(from_meta_path, to_meta_path))
        return True

    def count(self, meta: CtxMeta) -> int:
        """
        Count attachments for meta

        :param meta: CtxMeta instance
        :return: number of attachments
        """
        return len(meta.get_additional_ctx())

    def delete(
            self,
            meta: CtxMeta,
            item: Dict[str, Any],
            delete_files: bool = False
    ):
        """
        Delete attachment

        :param meta: CtxMeta instance
        :param item: Attachment item dict
        :param delete_files: delete files
        """
        meta.remove_additional_ctx(item)
        self.window.core.ctx.save(meta.id)
        if delete_files:
            self.delete_local(meta, item)
        if len(meta.get_additional_ctx()) == 0:
            self.delete_index(meta)

    def delete_by_meta(self, meta: CtxMeta):
        """
        Delete all attachments for meta

        :param meta: CtxMeta instance
        """
        self.delete_index(meta)

    def delete_by_meta_id(self, meta_id: int):
        """
        Delete all attachments for meta by id

        :param meta_id: Meta id
        """
        meta = self.window.core.ctx.get_meta_by_id(meta_id)
        if meta is not None:
            self.delete_index(meta)

    def reset_by_meta(
            self,
            meta: CtxMeta,
            delete_files: bool = False
    ):
        """
        Delete all attachments for meta

        :param meta: CtxMeta instance
        :param delete_files: delete files
        """
        meta.reset_additional_ctx()
        self.window.core.ctx.save(meta.id)
        if delete_files:
            self.delete_index(meta)

    def reset_by_meta_id(
            self,
            meta_id: int,
            delete_files: bool = False
    ):
        """
        Delete all attachments for meta by id

        :param meta_id: Meta id
        :param delete_files: delete files
        """
        meta = self.window.core.ctx.get_meta_by_id(meta_id)
        if meta is not None:
            self.reset_by_meta(meta, delete_files)

    def clear(
            self,
            meta: CtxMeta,
            delete_files: bool = False
    ):
        """
        Clear all attachments by ctx meta

        :param meta: CtxMeta instance
        :param delete_files: delete files
        """
        meta.reset_additional_ctx()
        self.window.core.ctx.save(meta.id)
        if delete_files:
            self.delete_index(meta)

    def delete_index(self, meta: CtxMeta):
        """
        Delete index by ctx meta

        :param meta: CtxMeta instance
        """
        meta_path = self.get_dir(meta)
        if os.path.exists(meta_path) and os.path.isdir(meta_path):
            shutil.rmtree(meta_path)
            if self.is_verbose():
                print("Attachment deleted dir: {}".format(meta_path))

    def delete_local(
            self,
            meta: CtxMeta,
            item: Dict[str, Any]
    ):
        """
        Delete local attachment

        :param meta: CtxMeta instance
        :param item: additional context item
        """
        file_id = item["uuid"]
        meta_path = self.get_dir(meta)
        index_path = os.path.join(meta_path, self.dir_index)
        file_idx_path = os.path.join(meta_path, file_id)
        if item["type"] == "local_file":
            if os.path.exists(file_idx_path):
                for f in os.listdir(file_idx_path):
                    os.remove(os.path.join(file_idx_path, f))
                    if self.is_verbose():
                        print("Attachment deleted: {}".format(f))
                os.rmdir(file_idx_path)

        # delete from index
        if "doc_ids" in item and isinstance(item["doc_ids"], list):
            for doc_id in item["doc_ids"]:
                self.window.core.idx.indexing.remove_attachment(index_path, doc_id)
                if self.is_verbose():
                    print("Attachment deleted doc ID: {}".format(doc_id))

    def truncate(self):
        """Truncate all attachments"""
        try:
            idx_path = self.window.core.config.get_user_dir("ctx_idx")
            if os.path.exists(idx_path) and os.path.isdir(idx_path):
                shutil.rmtree(idx_path)
            os.makedirs(idx_path, exist_ok=True)
        except Exception as e:
            self.window.core.debug.error("Attachment.truncate", e)

    def reset(self):
        """Reset context info"""
        self.last_used_item = None
        self.last_used_content = None
        self.last_used_context = None
        self.last_files = []
        self.last_urls = []

    def get_used_files(self) -> List[str]:
        """
        Get last used files

        :return: list of files
        """
        return self.last_files

    def get_used_urls(self) -> List[str]:
        """
        Get last used URLs

        :return: list of URLs
        """
        return self.last_urls

    def is_verbose(self) -> bool:
        """
        Check if verbose mode is enabled

        :return: True if verbose mode is enabled
        """
        return self.window.core.config.get("ctx.attachment.verbose", False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from PySide6.QtCore import QEvent
from PySide6.QtWidgets import QTabWidget, QWidget

from pygpt_net.core.tabs.tab import Tab


class TabBody(QTabWidget):
    def __init__(self, window=None):
        super(TabBody, self).__init__()
        self.window = window
        self.owner = None
        self.body = []
        self.installEventFilter(self)

    def append(self, body: QWidget):
        """
        Append tab body (parent widget)

        :param body: tab body widget
        """
        self.body.append(body)

    def get_body(self):
        """
        Get tab body (parent widget)

        :return: tab body widget
        """
        return self.body

    def attach_tab(self, tab: Tab):
        """
        Attach tab to body

        :param tab: tab instance
        """
        for body in self.body:
            if hasattr(body, 'set_tab'):
                body.set_tab(tab)

    def setOwner(self, owner: Tab):
        """
        Set tab parent (owner)

        :param owner: parent tab instance
        """
        self.owner = owner
        self.attach_tab(owner)

    def getOwner(self) -> Tab:
        """
        Get tab parent (owner)

        :return: parent tab instance
        """
        return self.owner

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if (event.type() == QEvent.ChildAdded and
                source is self and
                event.child().isWidgetType()):
            self._glwidget = event.child()
            self._glwidget.installEventFilter(self)
        elif (event.type() == event.Type.MouseButtonPress):
            # handle column focus
            if self.owner is not None:
                col_idx = self.owner.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.17 01:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QWidget

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans
from .widgets import IdxBrowser


class BrowseTab:
    def __init__(self, window=None):
        """
        Tab: Browse

        :param window: Window instance
        """
        self.window = window
        self.browser = None

    def setup(self):
        """
        Setup tab widget
        """
        # idx db browser
        self.browser = IdxBrowser(self.window)
        self.window.ui.nodes['tool.indexer.browser'] = self.browser

        # layout
        self.window.ui.nodes["tool.indexer.browse.header.tip"] = HelpLabel(trans("tool.indexer.tab.browse.tip"))
        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes["tool.indexer.browse.header.tip"])
        layout.addWidget(self.browser)

        widget = QWidget()
        widget.setLayout(layout)

        return widget



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 22:00:00                  #
# ================================================== #

from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.ui.widget.element.group import CollapsedGroup
from pygpt_net.ui.widget.option.checkbox import OptionCheckbox
from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.ui.widget.option.textarea import OptionTextarea
from pygpt_net.utils import trans

import pygpt_net.icons_rc


class OptionCmd(QWidget):
    def __init__(
            self,
            window=None,
            plugin: BasePlugin = None,
            parent_id: str = None,
            id: str = None,
            option: dict = None
    ):
        """
        Command dictionary option widget

        :param window: main window
        :param plugin: plugin instance
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(OptionCmd, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option  # option data

        # TODO: implement cmd type option
        # print(self.option)

        key = option['id']
        cmd_id = option['id'].replace("cmd.", "")
        txt_desc = option['description']
        txt_tooltip = option['tooltip']

        # locale
        if plugin.use_locale:
            domain = 'plugin.' + plugin.id
            txt_desc = trans(key + '.description', False, domain)
            txt_tooltip = trans(key + '.tooltip', False, domain)
            # check if translation exists
            if txt_desc == key + '.description':
                txt_desc = trans("settings.cmd.field.desc").format(cmd=cmd_id)
            if txt_tooltip == key + '.tooltip':
                txt_tooltip = trans("settings.cmd.field.tooltip").format(cmd=cmd_id)

        # enable
        option_enabled = {}
        option_enabled["type"] = "bool"
        option_enabled["label"] = trans("settings.cmd.field.enable").format(cmd=cmd_id)
        option_enabled["description"] = ""
        option_enabled["value"] = True

        # params
        option_params = {}
        option_params["name"] = "params"
        option_params["type"] = "dict"
        option_params["keys"] = self.option["params_keys"]
        option_params["value"] = []
        if self.option["value"] is not None and "params" in self.option["value"]:
            option_params["value"] = self.option["value"]["params"]
        option_params["tooltip"] = ""

        # instruction
        option_instruction = {}
        option_instruction["type"] = "textarea"
        option_instruction["label"] = "Instruction"
        option_instruction["value"] = ""

        # keys
        key_enabled = self.id + ".enabled"
        key_params = self.id + ".params"
        key_instruction = self.id + ".instruction"

        label_key = self.parent_id + '.' + id + '.label'
        desc_key = self.parent_id + '.' + id + '.desc'

        self.window.ui.nodes[desc_key] = QLabel(txt_desc)
        self.window.ui.nodes[desc_key].setWordWrap(True)
        self.window.ui.nodes[desc_key].setMaximumHeight(40)
        self.window.ui.nodes[desc_key].setStyleSheet("font-size: 10px;")
        self.window.ui.nodes[desc_key].setProperty('class', 'label-help')
        self.window.ui.nodes[desc_key].setContentsMargins(35, 0, 0, 0)

        instr_key = "settings.cmd.field.instruction"
        params_key = "settings.cmd.field.params"
        instr_label = QLabel(trans(instr_key))
        params_label = QLabel(trans(params_key))

        # widgets
        self.enabled = OptionCheckbox(self.window, parent_id, key_enabled, option_enabled, icon = ":/icons/build.svg")  # enable checkbox
        self.params = OptionDict(self.window, parent_id, key_params, option_params)  # command params
        self.instruction = OptionTextarea(self.window, parent_id, key_instruction, option_instruction)  # command instruction

        # layout
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.enabled)
        self.layout.addWidget(self.window.ui.nodes[desc_key])

        params_layout = QVBoxLayout()
        params_layout.addWidget(instr_label)
        params_layout.addWidget(self.instruction)
        params_layout.addWidget(params_label)
        params_layout.addWidget(self.params)

        group_id = self.parent_id + '.' + id + '.config'
        group = CollapsedGroup(self.window, group_id, None, False, None)
        group.box.setText(trans('settings.cmd.config.collapse'))
        group.box.setIcon(QIcon(":/icons/expand.svg"))
        group.add_layout(params_layout)
        group.layout.setContentsMargins(25, 0, 0, 0)

        # add to groups
        self.window.ui.groups[group_id] = group

        self.layout.addWidget(group)
        self.setLayout(self.layout)

        # show tooltip only if different from description
        # if txt_tooltip != txt_desc:
            # self.setToolTip(txt_tooltip)

        # update
        self.update()

    def update_item(self, idx, data):
        """
        Update item in params list

        :param idx: Item index
        :param data: Item data
        """
        self.params.update_item(idx, data)

    def update(self):
        """Update widget"""
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 04:00:00                  #
# ================================================== #

class PidData():

    def __init__(self, pid, meta=None):
        """Pid Data"""
        self.pid = pid
        self.meta = meta
        self.images_appended = []
        self.urls_appended = []
        self.files_appended = []
        self.buffer = ""
        self.prev_position = None  # previous cursor position (for chunk append)
        self.is_cmd = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.17 17:00:00                  #
# ================================================== #

from pygpt_net.core.docker import Docker as BaseDocker
from pygpt_net.core.docker.builder import Builder


class Docker(BaseDocker):
    def __init__(self, plugin = None):
        super(Docker, self).__init__(plugin)
        self.plugin = plugin
        self.initialized = False
        self.signals = None
        self.builder = Builder(self.plugin)
        self.builder.docker = self

    def build(self):
        """Run image build"""
        self.builder.build_image()

    def build_and_restart(self):
        """Run image build and restart container"""
        self.builder.build_image(restart=True)

    def get_dockerfile(self) -> str:
        """
        Get the Dockerfile

        :return: Dockerfile.
        """
        return self.plugin.get_option_value('dockerfile')

    def get_image_name(self) -> str:
        """
        Get the image name

        :return: Image name.
        """
        return self.plugin.get_option_value('image_name')

    def get_container_name(self) -> str:
        """
        Get the container name

        :return: Container name.
        """
        return self.plugin.get_option_value('container_name')

    def get_local_data_dir(self) -> str:
        """
        Get the local data directory.

        :return: Local data directory.
        """
        return self.plugin.window.core.config.get_user_dir("data")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, Any

from pygpt_net.utils import trans


class Cmd:
    def __init__(self, window=None):
        """
        Command field handler - >>> NOT IMPLEMENTED YET <<<<

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ):
        """
        Apply values to command fields

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        if option["value"] is None:
            option["value"] = {
                "enabled": False,
                "instruction": "",
                "params": [],
            }
        if "params" not in option["value"]:
            option["value"]["params"] = []
        if "enabled" not in option["value"]:
            option["value"]["enabled"] = False
        if "instruction" not in option["value"]:
            option["value"]["instruction"] = ""

        params = list(option["value"]["params"])
        self.window.ui.config[parent_id][key].params.items = params  # replace model data list
        self.window.ui.config[parent_id][key].params.model.updateData(params)  # update model data
        self.window.ui.config[parent_id][key].enabled.box.setChecked(option["value"]["enabled"])
        self.window.ui.config[parent_id][key].instruction.setText(str(option["value"]["instruction"]))

    def apply_row(
            self,
            parent_id: str,
            key: str,
            values: Dict[str, Any],
            idx: int
    ):
        """
        Apply data values to command fields

        :param parent_id: Options parent ID
        :param key: Option key
        :param values: dictionary data values
        :param idx: row index
        """
        # remove .params suffix
        if key.endswith(".params"):
            key = key.replace(".params", "")
        self.window.ui.config[parent_id][key].params.update_item(idx, values["params"])

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ) -> dict:
        """
        Get command values

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        :return: list with dictionary values
        """
        data = {
            "enabled": self.window.ui.config[parent_id][key].enabled.box.isChecked(),
            "instruction": self.window.ui.config[parent_id][key].instruction.toPlainText(),
            "params": self.window.ui.config[parent_id][key].params.model.items,
        }
        return data

    def to_options(
            self,
            parent_id: str,
            option: Dict[str, Any]
    ) -> dict:
        """
        Convert dictionary items option to options

        :param parent_id: Parent ID
        :param option: dictionary items option
        :return: options dict
        """
        # option type: dict
        if option["type"] == "dict":
            if "keys" not in option:
                return {}
            options = {}
            for key in option["keys"]:
                item = option["keys"][key]
                if isinstance(item, str):
                    item = {
                        "label": parent_id + '.' + key,
                        "type": item,  # field type is provided as value in this case
                    }
                options[key] = item
                if "label" not in options[key]:
                    options[key]["label"] = key
                    options[key]["label"] = parent_id + "." + options[key]["label"]
            return options

        # option type: cmd
        elif option["type"] == "cmd":
            if "params_keys" not in option:  # keys are stored in "params_keys" in cmd type
                return {}
            options = {}
            for key in option["params_keys"]:
                item = option["params_keys"][key]
                if isinstance(item, str):
                    item = {
                        "label": 'dictionary.cmd.param.' + key,
                        "type": item,  # field type is provided as value in this case
                    }
                options[key] = item
                if "label" not in options[key]:
                    options[key]["label"] = 'dictionary.cmd.param.' + key
            return options



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any, Optional

from packaging.version import Version

from pygpt_net.item.index import IndexItem


class BaseProvider:
    def __init__(self, window=None):
        self.window = window
        self.id = ""
        self.type = "index"

    def attach(self, window):
        self.window = window

    def install(self):
        pass

    def patch(self, version: Version) -> bool:
        pass

    def load(
            self,
            store_id: str
    ) -> Dict[str, IndexItem]:
        pass

    def append_file(
            self,
            store_id: str,
            idx: str,
            data: Dict[str, Any]
    ) -> int:
        pass

    def append_ctx_meta(
            self,
            store_id: str,
            idx: str,
            meta_id: int,
            doc_id: str
    ) -> int:
        pass

    def append_external(
            self,
            store_id: str,
            idx: str,
            data: Dict[str, Any]
    ) -> int:
        pass

    def update_file(
            self,
            id: int,
            doc_id: str,
            ts: int
    ) -> bool:
        pass

    def update_ctx_meta(
            self,
            id: int,
            doc_id: str
    ) -> bool:
        pass

    def update_external(
            self,
            content: str,
            type: str,
            doc_id: str,
            ts: int
    ) -> bool:
        pass

    def is_meta_indexed(
            self,
            store_id: str,
            idx: str,
            meta_id: int
    ) -> bool:
        pass

    def is_file_indexed(
            self,
            store_id: str,
            idx: str,
            file_id: str
    ) -> bool:
        pass

    def is_external_indexed(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str
    ) -> bool:
        pass

    def get_meta_doc_id(
            self,
            store_id: str,
            idx: str,
            meta_id: int
    ) -> str:
        pass

    def get_file_doc_id(
            self,
            store_id: str,
            idx: str,
            file_id: str
    ) -> str:
        pass

    def get_external_doc_id(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str
    ) -> str:
        pass

    def remove_file(
            self,
            store_id: str,
            idx: str,
            doc_id: str
    ):
        pass

    def remove_external(
            self,
            store_id: str,
            idx: str,
            doc_id: str
    ):
        pass

    def remove_ctx_meta(
            self,
            store_id: str,
            idx: str,
            meta_id: str
    ):
        pass

    def truncate(
            self,
            store_id: str,
            idx: str
    ):
        pass

    def truncate_all(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        pass

    def truncate_files(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        pass

    def truncate_ctx(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        pass

    def truncate_external(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        pass

    def get_counters(
            self,
            type: str
    ) -> Dict[str, Dict[str, int]]:
        pass

    def dump(self, index: IndexItem) -> str:
        pass

    def get_version(self) -> str:
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 04:00:00                  #
# ================================================== #

import datetime
import smtplib
import poplib

from email.parser import BytesParser
from typing import Any
from bs4 import BeautifulSoup

from pygpt_net.item.ctx import CtxItem


class Runner:
    def __init__(self, plugin=None):
        """
        Cmd Runner

        :param plugin: plugin
        """
        self.plugin = plugin
        self.signals = None

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    def parse_email(self, msg, as_text: bool = True) -> str:
        """
        Parse email message

        :param msg: email message
        :param as_text: return as text
        :return: parsed email
        """
        body = ""
        is_html = False
        if msg.is_multipart():
            for part in msg.walk():
                content_type = part.get_content_type()
                content_disposition = str(part.get('Content-Disposition'))
                if content_type == 'text/plain' and 'attachment' not in content_disposition:
                    body = part.get_payload(decode=True).decode(part.get_content_charset('utf-8'))
                elif content_type == 'text/html' and 'attachment' not in content_disposition:
                    body = part.get_payload(decode=True).decode(part.get_content_charset('utf-8'))
                    is_html = True
        else:
            body = msg.get_payload(decode=True).decode(msg.get_content_charset('utf-8'))
        if is_html and as_text:
            body = BeautifulSoup(body, 'html.parser').get_text()
        return body

    def smtp_get_email_body(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Receive emails from mailbox

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        server = self.plugin.get_option_value("smtp_host")
        port = self.plugin.get_option_value("smtp_port_inbox")
        id = item["params"].get("id", "")
        username = self.plugin.get_option_value("smtp_user")
        password = self.plugin.get_option_value("smtp_password")
        msg = "Receiving email from: {}".format(server)
        format = "text"
        as_text = True
        if "format" in item["params"]:
            format = item["params"]["format"].lower()
        if format == "html":
            as_text = False
        self.log(msg)
        try:
            if port in [995]:
                pop3 = poplib.POP3_SSL(server, port)
            else:
                pop3 = poplib.POP3(server, port)

            msg = pop3.getwelcome().decode('utf-8')
            self.log(msg)
            pop3.user(username)
            pop3.pass_(password)
            response, lines, octets = pop3.retr(id)
            msg_content = b'\r\n'.join(lines)
            msg = BytesParser().parsebytes(msg_content)
            result = {
                "ID": id,
                "From": msg["From"],
                "Subject": msg["Subject"],
                "Date": msg["Date"],
                "Message": self.parse_email(msg, as_text)
            }
            pop3.quit()
        except Exception as e:
            self.error(e)
            result = "Error: {}".format(str(e))
        return {
            "request": request,
            "result": str(result)
        }

    def smtp_receive_emails(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Receive emails from mailbox

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        server = self.plugin.get_option_value("smtp_host")
        port = self.plugin.get_option_value("smtp_port_inbox")
        limit = item["params"].get("limit", 10)
        offset = item["params"].get("offset", 0)
        username = self.plugin.get_option_value("smtp_user")
        password = self.plugin.get_option_value("smtp_password")
        msg = "Receiving emails from: {}".format(server)
        order = "desc"
        if "order" in item["params"]:
            tmp_order = item["params"]["order"].lower()
            if tmp_order in ["asc", "desc"]:
                order = tmp_order
        self.log(msg)

        messages = []
        try:
            if port in [995]:
                pop3 = poplib.POP3_SSL(server, port)
            else:
                pop3 = poplib.POP3(server, port)

            msg = pop3.getwelcome().decode('utf-8')
            self.log(msg)
            pop3.user(username)
            pop3.pass_(password)

            message_count, mailbox_size = pop3.stat()
            total = message_count
            print(f'Num of messages: {message_count}, Inbox size: {mailbox_size} bytes')

            if order == "desc":
                if offset > 0:
                    message_count -= offset
                if message_count < limit:
                    limit = message_count
                range_max = message_count - limit
                if limit == 0:
                    range_max = 0
                for i in range(message_count, range_max, -1):
                    response, lines, octets = pop3.top(i, 0)
                    msg_content = b'\r\n'.join(lines)
                    msg = BytesParser().parsebytes(msg_content)
                    msg_body = {
                        "ID": i,
                        "From": msg["From"],
                        "Subject": msg["Subject"],
                        "Date": msg["Date"],
                    }
                    messages.append(msg_body)
            else:
                if limit == 0:
                    limit = message_count
                else:
                    limit += offset
                if limit > message_count:
                    limit = message_count
                for i in range(offset + 1, limit + 1):
                    response, lines, octets = pop3.top(i, 0)
                    msg_content = b'\r\n'.join(lines)
                    msg = BytesParser().parsebytes(msg_content)
                    msg_body = {
                        "ID": i,
                        "From": msg["From"],
                        "Subject": msg["Subject"],
                        "Date": msg["Date"],
                    }
                    messages.append(msg_body)

            pop3.quit()
            result = {
                "total": total,
                "messages": messages,
            }
        except Exception as e:
            self.error(e)
            result = "Error: {}".format(str(e))
        return {
            "request": request,
            "result": str(result)
        }

    def smtp_send_email(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command on host

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        msg = "Sending email to: {}".format(item["params"]['recipient'])
        self.log(msg)
        try:
            email_to = item["params"]['recipient']
            message = item["params"]['message']
            subject = item["params"]['subject']
            from_addr = self.plugin.get_option_value("from_email")
            host = self.plugin.get_option_value("smtp_host")
            port = self.plugin.get_option_value("smtp_port_outbox")
            user = self.plugin.get_option_value("smtp_user")
            password = self.plugin.get_option_value("smtp_password")
            date = datetime.datetime.now().strftime("%d/%m/%Y %H:%M")
            body = """From: {sender}\nTo: {to}\nDate: {date}\nSubject: {subject}\n\n{msg}
                    """.format(sender=from_addr, to=email_to, date=date, subject=subject, msg=message)
            if port in [465]:
                server = smtplib.SMTP_SSL(host, port)
            else:
                server = smtplib.SMTP(host, port)
            server.ehlo()
            if port in [587]:
                server.starttls()
            server.login(user, password)
            server.sendmail(from_addr, email_to, body)
            server.close()
            result = "OK"
        except Exception as e:
            self.error(e)
            result = "Error: {}".format(str(e))
        return {
            "request": request,
            "result": str(result)
        }

    def error(self, err: Any):
        """
        Log error message

        :param err: exception or error message
        """
        if self.signals is not None:
            self.signals.error.emit(err)

    def status(self, msg: str):
        """
        Send status message

        :param msg: status message
        """
        if self.signals is not None:
            self.signals.status.emit(msg)

    def debug(self, msg: Any):
        """
        Log debug message

        :param msg: message to log
        """
        if self.signals is not None:
            self.signals.debug.emit(msg)

    def log(self, msg, sandbox: bool = False):
        """
        Log message to console

        :param msg: message to log
        :param sandbox: is sandbox mode
        """
        full_msg = '[SMTP]' + ' ' + str(msg)
        if self.signals is not None:
            self.signals.log.emit(full_msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, Any

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_EXPERT,
)
from pygpt_net.core.events import KernelEvent, RenderEvent
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.ctx.reply import ReplyContext
from pygpt_net.item.ctx import CtxItem


class Experts:
    def __init__(self, window=None):
        """
        Experts controller

        :param window: Window instance
        """
        self.window = window
        self.is_stop = False

    def stop(self):
        """
        Stop experts
        """
        self.is_stop = True

    def stopped(self) -> bool:
        """
        Check if experts are stopped

        :return: True if experts are stopped
        """
        return self.is_stop

    def unlock(self):
        """Unlock experts"""
        self.is_stop = False

    def enabled(self) -> bool:
        """
        Check if experts are enabled

        :return: True if experts are enabled
        """
        modes = [MODE_AGENT, MODE_EXPERT]
        mode = self.window.core.config.get('mode')
        if mode in modes or self.window.controller.plugins.is_type_enabled("expert"):
            return True
        return False

    def append_prompts(
            self,
            mode: str,
            sys_prompt: str,
            parent_id: Optional[str] = None
    ):
        """
        Append prompt to the window

        :param mode: Mode
        :param sys_prompt: Prompt text
        :param parent_id: Parent ID
        """
        # if agent enabled
        if self.window.controller.agent.legacy.enabled():
            prev_prompt = sys_prompt
            sys_prompt = self.window.core.prompt.get("agent.instruction")
            if prev_prompt is not None and prev_prompt.strip() != "":
                sys_prompt = sys_prompt + "\n\n" + prev_prompt  # append previous prompt

        # expert or agent mode
        if self.window.controller.agent.experts.enabled() and parent_id is None:  # master expert has special prompt
            if self.window.controller.agent.legacy.enabled():  # if agent then leave agent prompt
                sys_prompt += "\n\n" + self.window.core.experts.get_prompt()  # both, agent + experts
            else:
                sys_prompt = self.window.core.experts.get_prompt()
                # mode = "chat"  # change mode to chat for expert

        # if global mode is agent
        if mode == MODE_AGENT:
            sys_prompt = self.window.controller.agent.legacy.on_system_prompt(
                sys_prompt,
                append_prompt=None,  # sys prompt from preset is used here
                auto_stop=self.window.core.config.get('agent.auto_stop'),
            )

        return sys_prompt

    def handle(self, ctx: CtxItem) -> int:
        """
        Handle mentions (calls) to experts

        :param ctx: CtxItem
        """
        stream_mode = self.window.core.config.get('stream')
        num_calls = 0

        # extract expert mentions
        if self.window.controller.agent.experts.enabled():
            # re-send to master
            if ctx.sub_reply:
                self.window.core.ctx.update_item(ctx)
                self.window.core.experts.reply(ctx)
            else:
                # call experts
                if not ctx.reply:
                    mentions = self.window.core.experts.extract_calls(ctx)
                    if mentions:
                        num_calls = 0
                        self.log("Calling experts...")
                        data = {
                            "meta": ctx.meta,
                            "ctx": ctx,
                            "stream": stream_mode,
                        }
                        event = RenderEvent(RenderEvent.END, data)
                        self.window.dispatch(event)  # close previous render
                        for expert_id in mentions:
                            if not self.window.core.experts.exists(expert_id):
                                self.log("Expert not found: " + expert_id)
                                continue
                            self.log("Calling: " + expert_id)
                            ctx.sub_calls += 1

                            # add to reply stack
                            reply = ReplyContext()
                            reply.type = ReplyContext.EXPERT_CALL
                            reply.ctx = ctx
                            reply.parent_id = expert_id
                            reply.input = mentions[expert_id]
                            # send to kernel
                            context = BridgeContext()
                            context.ctx = ctx
                            context.reply_context = reply
                            event = KernelEvent(KernelEvent.AGENT_CALL, {
                                'context': context,
                                'extra': {},
                            })
                            self.window.dispatch(event)

                            num_calls += 1
                        if num_calls > 0:
                            return num_calls  # abort continue if expert call detected
        return num_calls

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.core.debug.info(data)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 03:00:00                  #
# ================================================== #

import time
import pyaudio
import wave
import numpy as np
import io
from pydub import AudioSegment

from PySide6.QtCore import Slot, Signal

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    volume_changed = Signal(float)  # Emits the volume level
    playback = Signal(str)
    stop = Signal()
    error_playback = Signal(object)


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.text = None
        self.event = None
        self.cache_file = None  # path to cache file
        self.mode = "generate"  # generate|playback
        self.audio_file = None
        self.is_stopped = False
        self.p = None
        self.stream = None

    @Slot()
    def run(self):
        try:
            if self.mode == "generate":
                self.generate()
            elif self.mode == "playback":
                self.play()
        except Exception as e:
            self.error(e)

    def generate(self):
        """Generate and play audio file"""
        if self.text is None or self.text == "":
            time.sleep(0.2)  # wait
            return
        path = self.plugin.get_provider().speech(self.text)
        if path:
            self.play_audio(path)
            # store in cache if enabled
            if self.cache_file:
                self.cache_audio_file(path, self.cache_file)

    def play(self):
        """Play audio file only"""
        if self.audio_file:
            self.play_audio(self.audio_file)

    def play_audio(self, audio_file: str):
        """
        Play audio file using PyAudio

        :param audio_file: audio file path
        """
        try:
            self.signals.playback.emit(self.event)
            audio = AudioSegment.from_file(audio_file)
            audio = audio.set_frame_rate(44100)  # resample to 44.1 kHz
            wav_io = io.BytesIO()
            audio.export(wav_io, format='wav')
            wav_io.seek(0)
            wf = wave.open(wav_io, 'rb')
            self.p = pyaudio.PyAudio()
            self.stream = self.p.open(format=self.p.get_format_from_width(wf.getsampwidth()),
                                      channels=wf.getnchannels(),
                                      rate=wf.getframerate(),
                                      output=True)

            sample_width = wf.getsampwidth()
            format = self.p.get_format_from_width(sample_width)

            if format == pyaudio.paInt8:
                dtype = np.int8
                max_value = 2 ** 7 - 1  # 127
                offset = 0
            elif format == pyaudio.paInt16:
                dtype = np.int16
                max_value = 2 ** 15 - 1  # 32767
                offset = 0
            elif format == pyaudio.paInt32:
                dtype = np.int32
                max_value = 2 ** 31 - 1  # 2147483647
                offset = 0
            elif format == pyaudio.paUInt8:
                dtype = np.uint8
                max_value = 2 ** 8 - 1  # 255
                offset = 128  # center unsigned data
            else:
                raise ValueError(f"Unsupported format: {format}")

            chunk_size = 512
            data = wf.readframes(chunk_size)

            while data != b'' and not self.is_stopped:
                self.stream.write(data)

                audio_data = np.frombuffer(data, dtype=dtype)
                if len(audio_data) > 0:
                    audio_data = audio_data.astype(np.float32)
                    if dtype == np.uint8:
                        audio_data -= offset

                    # compute RMS
                    rms = np.sqrt(np.mean(audio_data ** 2))

                    if rms > 0:
                        # RMS to decibels
                        db = 20 * np.log10(rms / max_value)

                        # define minimum and maximum dB levels
                        min_db = -60  # adjust as needed
                        max_db = 0

                        # clamp the db value to the range [min_db, max_db]
                        if db < min_db:
                            db = min_db
                        elif db > max_db:
                            db = max_db

                        # map decibel value to volume percentage
                        volume_percentage = ((db - min_db) / (max_db - min_db)) * 100
                    else:
                        volume_percentage = 0

                    # emit volume signal
                    self.signals.volume_changed.emit(volume_percentage)
                else:
                    # if empty audio_data
                    self.signals.volume_changed.emit(0)

                data = wf.readframes(chunk_size)

            # close the stream
            if self.stream is not None:
                if self.stream.is_active():
                    self.stream.stop_stream()
                self.stream.close()
            if self.p is not None:
                self.p.terminate()
            wf.close()
            self.signals.volume_changed.emit(0)
        except Exception as e:
            self.signals.volume_changed.emit(0)
            self.signals.error_playback.emit(e)

    def cache_audio_file(self, src: str, dst: str):
        """
        Store audio file in cache

        :param src: source path
        :param dst: destination path
        """
        import shutil
        try:
            shutil.copy(src, dst)
            # print("Cached audio file:", dst)
        except Exception as e:
            self.error(e)

    def send(self, playback):
        """
        Send playback object to main thread

        :param playback: playback object
        """
        self.signals.playback.emit(playback, self.event)

    def stop_playback(self):
        """Stop audio playback"""
        self.stop()

    def stop(self):
        """Send stop signal to main thread"""
        self.is_stopped = True
